from pathlib import Path

from .patch_util import patch_with_status_update


def patch(input_path: Path, output_path: Path, configuration: dict):
    from .dread_patcher import patch_extracted
    return patch_extracted(input_path, output_path, configuration)


__all__ = [
    patch,
    patch_with_status_update,
]



from open_dread_rando.misc_patches.tunable_patcher import apply_tunable_patches
from open_dread_rando.patcher_editor import PatcherEditor


def apply_cosmetic_patches(editor: PatcherEditor, cosmetic: dict):
    apply_tunable_patches(editor, cosmetic["config"])



from open_dread_rando.logger import LOG
from open_dread_rando.patcher_editor import PatcherEditor


def is_tilegroup(actor):
    return "TILEGROUP" in actor.pComponents and actor.pComponents.TILEGROUP["@type"] == "CBreakableTileGroupComponent"


def patch_tilegroup(editor: PatcherEditor, group: dict):
    """
    Patches a tilegroup from its original tile type into a new tile type

    :param editor: the PatcherEditor
    :param group: a dictionary containing the actor reference stored in 'actor' key and a tile type
    stored in 'tiletype' key
    """
    actor = editor.resolve_actor_reference(group["actor"])

    if not is_tilegroup(actor):
        raise ValueError(f"Actor at {group['actor']} is not a breakable tile group.")

    gridTiles = actor.pComponents.TILEGROUP.aGridTiles
    for tile in gridTiles:
        tile.eTileType = group["tiletype"]

def patch_individual_tiles(editor: PatcherEditor, actor_reference: dict, tiles: list):
    actor = editor.resolve_actor_reference(actor_reference)

    if not is_tilegroup(actor):
        raise ValueError(f"Actor at {actor_reference} is not a breakable tile group.")

    gridTiles = actor.pComponents.TILEGROUP.aGridTiles

    if len(tiles) != len(gridTiles):
        tiles_string = f"got {len(tiles)}, expected {len(gridTiles)}"

        if len(tiles) < len(gridTiles):
            raise ValueError(f"Too few tiles provided to patch tilegroup {actor_reference} ({tiles_string})")
        else:
            LOG.warning(f"More tiles provided than in tilegroup {actor_reference} ({tiles_string})")

    for i, tile in enumerate(gridTiles):
        tile.eTileType = tiles[i]["tiletype"]



from __future__ import annotations

import typing

from open_dread_rando.patcher_editor import PatcherEditor


class ActorReferenceTuple(typing.NamedTuple):
    scenario: str
    actor_layer: str
    sublayer: str
    actor: str

    def from_dict(reference: dict[str, str]) -> ActorReferenceTuple:
        return ActorReferenceTuple(
            reference["scenario"],
            reference["actor_layer"],
            reference.get("sublayer", reference.get("layer")),
            reference["actor"]
        )

def _remove_all_actors(editor: PatcherEditor, scenario_name: str, actor_layer: str) -> set[ActorReferenceTuple]:
    to_remove = set()
    scenario = editor.get_scenario(scenario_name)

    for sublayer_name, actor_name, actor in scenario.all_actors_in_actor_layer(actor_layer):
        to_remove.add(ActorReferenceTuple(scenario_name, actor_layer, sublayer_name, actor_name))

    return to_remove

def _remove_actors_from_groups(editor: PatcherEditor, scenario_name: str, actor_layer: str,
                               actor_groups: list[str]) -> set[ActorReferenceTuple]:
    to_remove = set()
    scenario = editor.get_scenario(scenario_name)

    for group in actor_groups:
        for actor_link in scenario.get_actor_group(group, actor_layer):
            to_remove.add(ActorReferenceTuple(**editor.reference_for_link(actor_link, scenario_name)))

    return to_remove

def _remove_actors_not_in_groups(editor: PatcherEditor, scenario_name: str, actor_layer: str,
                                 actor_groups: list[str]) -> tuple[set[ActorReferenceTuple], set[ActorReferenceTuple]]:
    to_remove = set()
    to_keep = set()
    scenario = editor.get_scenario(scenario_name)

    for group in scenario.actor_groups_for_actor_layer(actor_layer):
        if group not in actor_groups:
            for actor_link in scenario.get_actor_group(group, actor_layer):
                to_remove.add(ActorReferenceTuple(**editor.reference_for_link(actor_link, scenario_name)))
        else:
            for actor_link in scenario.get_actor_group(group, actor_layer):
                to_keep.add(ActorReferenceTuple(**editor.reference_for_link(actor_link, scenario_name)))

    return to_remove, to_keep

def mass_delete_actors(editor: PatcherEditor, configuration: dict) -> None:
    # Sets of tuples will be used to ensure no duplicate entries
    to_remove = set()
    to_keep = { ActorReferenceTuple.from_dict(reference) for reference in configuration["to_keep"] }

    for scenario_config in configuration["to_remove"]:
        scenario_name = scenario_config["scenario"]
        actor_layer = scenario_config["actor_layer"]
        method = scenario_config["method"]

        if method == "all":
            to_remove.update(_remove_all_actors(editor, scenario_name, actor_layer))

        elif method == "remove_from_groups":
            to_remove.update(_remove_actors_from_groups(editor, scenario_name, actor_layer,
                                                       scenario_config["actor_groups"]))

        elif method == "keep_from_groups":
            new_remove, new_keep = _remove_actors_not_in_groups(editor, scenario_name, actor_layer,
                                                                scenario_config["actor_groups"])

            to_remove.update(new_remove)
            to_keep.update(new_keep)

    to_remove.difference_update(to_keep)

    for actor in to_remove:
        editor.remove_entity(actor._asdict(), None)



import copy
from bisect import insort
from collections.abc import Sequence
from enum import Enum
from pathlib import Path

from construct import Container, ListContainer
from mercury_engine_data_structures.formats import Bmscc

from open_dread_rando.constants import ALL_SCENARIOS
from open_dread_rando.patcher_editor import PatcherEditor

# copied from existing entity, so we don't have to make a whole shield
_EXAMPLE_SHIELD = {"scenario": "s010_cave", "sublayer": "default", "actor": "Door003_missileShield"}


class MinimapIconData(Enum):
    """
    Enum representing minimap data. Also contains functions that return mapDoor and mapBlockage Containers.

    iconId: a non-directional icon id. 'L' or 'R' is appended for each direction.
    tuple: the offsets for the oBoxes. All values should be taken from a "left" entity:
        0: x-axis offset from vPos to the bottom-right corner
        1: x-axis offset from vPos to the top-left corner
        2: y-axis offset from vPos to the bottom-right corner
        3: y-axis offset from vPos to the top-left corner
    """
    DOOR_FRAME = ("DoorFrame", (-150, -50, 0, 300))
    DOOR_POWER = ("DoorPower", (-150, -50, 0, 300))
    DOOR_CHARGE = ("DoorCharge", (-150, -50, 0, 300))
    DOOR_GRAPPLE = ("DoorGrapple", (-150, -50, 0, 300))
    DOOR_PRESENCE = ("DoorPresence", (-150, -50, 0, 300))
    SHIELD_WIDE_BEAM = ("DoorWide", (-300, -150, 0, 300))
    SHIELD_PLASMA_BEAM = ("BlockagePlasma", (-300, -150, 0, 300))
    SHIELD_WAVE_BEAM = ("BlockageWave", (-300, -150, 0, 300))
    SHIELD_DIFFUSION = ("BlockageDiffusion", (-300, -150, 0, 300))
    SHIELD_MISSILE = ("BlockageMissile", (-300, -150, 0, 300))
    SHIELD_SUPER_MISSILE = ("BlockageSuperMissile", (-300, -150, 0, 300))
    SHIELD_ICE_MISSILE = ("BlockageIce", (-300, -150, 0, 300))
    SHIELD_STORM_MISSILE = ("BlockageStorm", (-300, -150, 0, 300))
    SHIELD_BOMB = ("BlockageBomb", (-300, -150, 0, 300))
    SHIELD_CROSS_BOMB = ("BlockageCrossBomb", (-300, -150, 0, 300))
    SHIELD_POWER_BOMB = ("BlockagePowerBomb", (-300, -150, 0, 300))
    SHIELD_CLOSED = ("BlockageClosed", (-300, -150, 0, 300))

    def __init__(self, icon_id: str, offsets: tuple[float, float, float, float]):
        """
        Initializes a new MinimapIconData with the given icon id and offsets for the left-facing direction
        """
        self.icon_id = icon_id
        self.oBox_min = (offsets[0], offsets[2])
        self.oBox_max = (offsets[1], offsets[3])

    def create_map_door(self, pos: Sequence[float]) -> Container:
        """creates a mapDoor for a given element"""
        cont = Container()
        cont["vPos"] = ListContainer([pos[0], pos[1]])
        cont["oBoxL"] = Container(
            Min=ListContainer([pos[0] + self.oBox_min[0], pos[1] - self.oBox_min[1]]),
            Max=ListContainer([pos[0] + self.oBox_max[0], pos[1] + self.oBox_max[1]]))
        cont["oBoxR"] = Container(
            Min=ListContainer([pos[0] - self.oBox_max[0], pos[1] - self.oBox_min[1]]),
            Max=ListContainer([pos[0] - self.oBox_min[0], pos[1] + self.oBox_max[1]]))
        cont["sLeftIconId"] = f"{self.icon_id}L"
        cont["sRightIconId"] = f"{self.icon_id}R"
        cont["aRoomIds"] = ListContainer()

        return cont

    def create_map_blockage(self, pos: Sequence[float], dir: str) -> Container:
        """Creates a mapBlockage for a shield in a given direction"""
        cont = Container()
        cont["vPos"] = ListContainer([pos[0], pos[1]])

        # get delta, which depends on shield dir.
        # shield L needs no adjustments so delta is zero.
        # shield R needs to be moved oBoxMin[0] + oBoxMax[0] to the right.
        delta = 0 if dir == "L" else -(self.oBox_min[0] + self.oBox_max[0])

        cont["oBox"] = Container(
            Min=ListContainer([pos[0] + self.oBox_min[0] + delta, pos[1] + self.oBox_min[1]]),
            Max=ListContainer([pos[0] + self.oBox_max[0] + delta, pos[1] + self.oBox_max[1]]))
        cont["sIconId"] = f"{self.icon_id}{dir}"
        cont["bFlipX"] = False if dir == "L" else True
        cont["bFlipY"] = False

        return cont


class ActorData(Enum):
    """
    Enum containing data on actors

    actordef: the actordef for the actor
    minimapData: the MinimapIconData for this type
    """
    DOOR_FRAME = (["doorframe"], MinimapIconData.DOOR_FRAME)
    DOOR_POWER = (["doorpowerpower", "doorpowerclosed", "doorclosedpower"], MinimapIconData.DOOR_POWER)
    DOOR_CHARGE = (["doorchargecharge", "doorchargeclosed", "doorclosedcharge"], MinimapIconData.DOOR_CHARGE)
    DOOR_GRAPPLE = (["doorgrapplegrapple", "doorgrappleclosed", "doorclosedgrapple"], MinimapIconData.DOOR_GRAPPLE)
    DOOR_PRESENCE = (["doorpresencepresence", "doorframepresence", "doorpresenceframe"], MinimapIconData.DOOR_PRESENCE)
    SHIELD_WIDE_BEAM = (["doorwidebeam"], MinimapIconData.SHIELD_WIDE_BEAM)
    SHIELD_PLASMA_BEAM = (["door_shield_plasma"], MinimapIconData.SHIELD_PLASMA_BEAM)
    SHIELD_WAVE_BEAM = (["doorwavebeam"], MinimapIconData.SHIELD_WAVE_BEAM)
    SHIELD_DIFFUSION_BEAM = (["shield_diffusion"], MinimapIconData.SHIELD_DIFFUSION)
    SHIELD_MISSILE = (["doorshieldmissile"], MinimapIconData.SHIELD_MISSILE)
    SHIELD_SUPER_MISSILE = (["doorshieldsupermissile"], MinimapIconData.SHIELD_SUPER_MISSILE)
    SHIELD_ICE_MISSILE = (["doorshieldicemissile"], MinimapIconData.SHIELD_ICE_MISSILE)
    SHIELD_STORM_MISSILE = (["doorshieldstormmissile"], MinimapIconData.SHIELD_STORM_MISSILE)
    SHIELD_BOMB = (["doorshieldbomb"], MinimapIconData.SHIELD_BOMB)
    SHIELD_CROSS_BOMB = (["doorshieldcrossbomb"], MinimapIconData.SHIELD_CROSS_BOMB)
    SHIELD_POWER_BOMB = (["doorshieldpowerbomb"], MinimapIconData.SHIELD_POWER_BOMB)
    SHIELD_CLOSED = (["doorshieldclosed"], MinimapIconData.SHIELD_CLOSED)

    def __init__(self, actordef: list[str], minimap: MinimapIconData):
        # generate actordefs
        self.actordefs = [f"actors/props/{v}/charclasses/{v}.bmsad" for v in actordef]
        self.minimapData = minimap


class DoorType(Enum):
    """
    Enum containing info on each door type

    type: the name Randovania calls the door
    door: the door's ActorData
    need_shield: whether the actor needs a shield
    shield: the shield's ActorData
    """
    FRAME = ("frame", ActorData.DOOR_FRAME)
    POWER = ("power_beam", ActorData.DOOR_POWER)
    CHARGE = ("charge_beam", ActorData.DOOR_CHARGE)
    DIFFUSION = ("diffusion_beam", ActorData.DOOR_POWER, True, ActorData.SHIELD_DIFFUSION_BEAM, True, True,
                 ["actors/props/door_shield_plasma"])
    WIDE_BEAM = ("wide_beam", ActorData.DOOR_POWER, True, ActorData.SHIELD_WIDE_BEAM, True, True,
                 ["actors/props/doorshieldmissile"])
    PLASMA_BEAM = ("plasma_beam", ActorData.DOOR_POWER, True, ActorData.SHIELD_PLASMA_BEAM)
    WAVE_BEAM = ("wave_beam", ActorData.DOOR_POWER, True, ActorData.SHIELD_WAVE_BEAM)
    MISSILE = ("missile", ActorData.DOOR_POWER, True, ActorData.SHIELD_MISSILE)
    SUPER_MISSILE = ("super_missile", ActorData.DOOR_POWER, True, ActorData.SHIELD_SUPER_MISSILE)
    SUPER_CHARGE_MISSILE = ("super_charge_missile", ActorData.DOOR_CHARGE, True,
                            ActorData.SHIELD_SUPER_MISSILE, True, False)
    ICE_MISSILE = ("ice_missile", ActorData.DOOR_POWER, True, ActorData.SHIELD_ICE_MISSILE, True, True,
                   ["actors/props/doorshieldmissile"])
    STORM_MISSILE = ("storm_missile", ActorData.DOOR_POWER, True, ActorData.SHIELD_STORM_MISSILE, True, True,
                     ["actors/props/doorshieldmissile", "actors/props/doorshieldsupermissile"])
    BOMB = ("bomb", ActorData.DOOR_POWER, True, ActorData.SHIELD_BOMB, True, True,
            ["actors/props/doorshieldmissile", "actors/props/doorshieldsupermissile"])
    CROSS_BOMB = ("cross_bomb", ActorData.DOOR_POWER, True, ActorData.SHIELD_CROSS_BOMB, True, True,
                  ["actors/props/doorshieldmissile", "actors/props/doorshieldsupermissile"])
    POWER_BOMB = ("power_bomb", ActorData.DOOR_POWER, True, ActorData.SHIELD_POWER_BOMB, True, True,
                  ["actors/props/doorshieldmissile", "actors/props/doorshieldsupermissile"])
    GRAPPLE = ("grapple_beam", ActorData.DOOR_GRAPPLE, False, None, True, True,
               ["actors/props/door"])
    PRESENCE = ("phantom_cloak", ActorData.DOOR_PRESENCE, False, None, True, False,
                ["actors/props/door"])
    CLOSED = ("closed", ActorData.DOOR_POWER, True, ActorData.SHIELD_CLOSED, True, True,
              ["actors/props/doorshieldmissile", "actors/props/doorshieldsupermissile"])

    def __init__(self, rdv_door_type: str, door_data: ActorData, need_shield: bool = False,
                 shield_data: ActorData = None, can_be_removed: bool = True, can_be_added: bool = True,
                 additional_asset_folders: list[str] = None):
        self.type = rdv_door_type
        self.need_shield = need_shield
        self.door = door_data
        self.shield = shield_data
        self.can_be_removed = can_be_removed
        self.can_be_added = can_be_added
        self.required_asset_folders = [] if additional_asset_folders is None else additional_asset_folders
        self.required_asset_folders.append(Path(self.door.actordefs[0]).parent.parent.as_posix())
        if self.need_shield:
            self.required_asset_folders.append(Path(self.shield.actordefs[0]).parent.parent.as_posix())

    @classmethod
    def get_type(cls, type: str):
        for e in cls:
            if e.type == type:
                return e

        raise ValueError(f"{type} is not a patchable door!")


def is_door(actor: Container):
    return "LIFE" in actor.pComponents and "CDoorLifeComponent" == actor.pComponents.LIFE["@type"]


class DoorPatcher:
    """An API to patch doors. Call patch_door() to use. """

    def __init__(self, editor: PatcherEditor) -> None:
        # get actors from reference dicts
        self.editor = editor
        self.available_shield_ids = {
            scenario: list(range(100))
            for scenario in ALL_SCENARIOS
        }
        self.SHIELD = editor.resolve_actor_reference(_EXAMPLE_SHIELD)
        self.rename_all_shields()
        self.patch_doorpresence_collision()

    def door_actor_to_type(self, door: Container, scenario: str) -> DoorType:
        """
        find's a door's DoorType
        creates a list of all doortype enums, and filters them by various means until there is one remaining
        raises an exception if door is not a patchable type
        """
        # make sure it's a door
        if not is_door(door):
            raise ValueError(f"Actor {door.sName} in scenario {scenario} is not a door!")

        # filter enums by door's actorref
        door_actor_ref = door.oActorDefLink.split(':')[1]
        possible_enum_values = list()
        for type in DoorType:
            for tdef in type.door.actordefs:
                if door_actor_ref == tdef:
                    possible_enum_values.append(type)

        if len(possible_enum_values) == 1:
            return possible_enum_values[0]

        # check for shields
        shield_actor = None
        life_comp = door.pComponents.LIFE
        for link_name in ["wpLeftDoorShieldEntity", "wpRightDoorShieldEntity"]:
            if life_comp[link_name] != '{EMPTY}':
                shield_actor = self.editor.resolve_actor_reference(
                    self.editor.reference_for_link(life_comp[link_name], scenario))

        if shield_actor is not None:
            # remove non-shielded doors if shield exists
            possible_enum_values = [e for e in possible_enum_values if e.need_shield]

            # check shields
            shield_actor_ref = shield_actor.oActorDefLink.split(':')[1]
            possible_enum_values = [e for e in possible_enum_values if e.shield.actordefs[0] == shield_actor_ref]
        else:
            # remove shielded doors if shield does not exist
            possible_enum_values = [e for e in possible_enum_values if not e.need_shield]

        if len(possible_enum_values) == 1:
            return possible_enum_values[0]
        else:
            raise ValueError(f"Door {door.sName} in scenario {scenario} is not a patchable door!")

    def patch_door(self, door_ref: dict, door_type: str):
        """
        Patches a door given a reference.

        @param door: A dictionary representing a requested door patch.
        """

        scenario = door_ref["scenario"]
        door_actor = self.editor.resolve_actor_reference(door_ref)

        # get the type of door we are patching to
        door_type = DoorType.get_type(door_type)
        if door_type.can_be_added is False:
            raise ValueError(f"Door type {door_type} cannot be patched in!")

        # get the type of door we are patching
        door_in_scenario_type = self.door_actor_to_type(door_actor, scenario)
        if door_in_scenario_type.can_be_removed is False:
            raise ValueError(
                f"Base game door {door_in_scenario_type.type} cannot be patched! "
                f"Requested door: {door_ref['actor']} in {scenario}"
            )

        self.door_to_basic(door_actor, door_in_scenario_type, scenario)
        self.power_to_door_type(door_actor, door_type, scenario)

    def door_to_basic(self, door: Container, door_type: DoorType, scenario: str):
        """
        Reverts a door to a basic (power) door based on its door_type
        """
        if door_type is DoorType.POWER:
            return

        if door_type.need_shield:
            self.remove_shields(door, scenario)

        # only needs patching if door isn't power
        if door_type.door is not ActorData.DOOR_POWER:
            self.any_door_to_power(door, scenario)

    # removes any shields if the door has them
    def remove_shields(self, door: Container, scenario: str):
        """
        Removes a door's shields. Assumes that the door has shields.

        param door: a door actor
        param scenario: the scenario string
        """
        life_comp = door.pComponents.LIFE
        for link_name in ["wpLeftDoorShieldEntity", "wpRightDoorShieldEntity"]:
            link = life_comp[link_name]
            if link == '{EMPTY}':
                continue  # skip shield if not linked

            # free the shield id if this is a rando shield
            shieldActor = self.editor.resolve_actor_reference(self.editor.reference_for_link(link, scenario))
            self.reclaim_old_shield_id(shieldActor.sName, scenario)
            self.editor.remove_entity(self.editor.reference_for_link(link, scenario), "mapBlockages")
            life_comp[link_name] = '{EMPTY}'

    # turns the door into a power beam door
    def any_door_to_power(self, door: Container, scenario: str):
        door.oActorDefLink = f"actordef:{ActorData.DOOR_POWER.actordefs[0]}"
        self.update_minimap_for_doors(door, DoorType.POWER, scenario)

    def power_to_door_type(self, door: Container, door_type: DoorType, scenario: str):
        # set door
        self.set_door_type(door, door_type, scenario)

        # if needed, set shield and ensure shield assets are present
        if door_type.need_shield:
            life_comp = door.pComponents["LIFE"]

            shield_l = self.create_shield(scenario, door, door_type.shield, "L")
            shield_r = self.create_shield(scenario, door, door_type.shield, "R")
            life_comp["wpLeftDoorShieldEntity"] = self.editor.build_link(shield_l.sName)
            life_comp["wpRightDoorShieldEntity"] = self.editor.build_link(shield_r.sName)

        # ensure assets are present
        for folder in door_type.required_asset_folders:
            for asset in self.editor.get_asset_names_in_folder(folder):
                self.editor.ensure_present_in_scenario(scenario, asset)

    def set_door_type(self, door: Container, door_type: DoorType, scenario: str):
        # set actor def to two-sided actordef
        door.oActorDefLink = f"actordef:{door_type.door.actordefs[0]}"
        self.update_minimap_for_doors(door, door_type, scenario)
        if door_type is DoorType.GRAPPLE:
            door.pComponents.LIFE.bStayOpen = False

    def create_shield(self, scenario: str, door: Container, shield_data: ActorData, dir: str):
        # make a new RandoShield by popping the lowest actor
        shield = self.editor.copy_actor(scenario, door.vPos, self.SHIELD, self.get_shield_id(scenario))

        self.editor.copy_actor_groups(
            { "actor": door.sName },
            { "actor": shield.sName},
            scenario
        )
        shield.oActorDefLink = f"actordef:{shield_data.actordefs[0]}"
        shield.vAng[1] = shield.vAng[1] if dir == "L" else -shield.vAng[1]
        if (shield_data is ActorData.SHIELD_WIDE_BEAM):
            shield.pComponents.LIFE['@type'] = 'CBeamDoorLifeComponent'

        self.update_minimap_for_shield(shield, shield_data, dir, scenario)
        return shield

    def update_minimap_for_doors(self, door: Container, door_type: DoorType, scenario: str):
        scenario_map = self.editor.get_scenario_map(scenario)
        map_doors = scenario_map.raw.Root.mapDoors
        patched_container = door_type.door.minimapData.create_map_door(door.vPos)
        map_doors[door.sName] = patched_container

    def update_minimap_for_shield(self, shield: Container, shield_type: ActorData, dir: str, scenario: str):
        scenario_map = self.editor.get_scenario_map(scenario)
        map_blockages = scenario_map.raw.Root.mapBlockages
        patched_container = shield_type.minimapData.create_map_blockage(shield.vPos, dir)
        map_blockages[shield.sName] = patched_container

    def remove_all_shields(self):
        for scenario in ALL_SCENARIOS:
            bmmap = self.editor.get_scenario_map(scenario)
            bmmap.raw.Root.mapBlockages = Container()

    def rename_all_shields(self):
        for scenario in ALL_SCENARIOS:
            brfld = self.editor.get_scenario(scenario)

            # we have to cache doors that have shields here and rename them outside the loop,
            # as otherwise it will rename actors in the actor list and confuse the program.
            shielded_doors = []
            for sublayer_name, actor_name, actor in list(brfld.all_actors_in_actor_layer()):

                # this is the door added to the Artaria CU.
                # For some reason is_door crashes on this so we add a check here.
                if actor_name == "DreadRando_CUDoor":
                    continue

                if not is_door(actor):
                    continue

                if actor.oActorDefLink[9:] not in ActorData.DOOR_POWER.actordefs:
                    continue

                shielded_doors.append(actor)

            for door in shielded_doors:
                self.rename_shields(door, scenario)

    def rename_shields(self, door: Container, scenario: str):
        life_comp = door.pComponents.LIFE
        for link_name in ["wpLeftDoorShieldEntity", "wpRightDoorShieldEntity"]:
            link = life_comp[link_name]
            if link == "{EMPTY}":
                continue

            # get shield actor and cache its sName
            reference = self.editor.reference_for_link(link, scenario)
            shieldActor = self.editor.resolve_actor_reference(reference)
            old_sName = shieldActor.sName

            # skip hdoors (doors where the environment covers one side of the door)
            # as they have terrain attached to the ShieldEntity links
            if "db_hdoor" in old_sName:
                continue

            # reclaim old shield id if this is a RandoShield
            self.reclaim_old_shield_id(shieldActor.sName, scenario)

            # grab the lowest open id
            new_id = self.get_shield_id(scenario)

            # make new actor, copy its groups, delete it
            brfld = self.editor.get_scenario(scenario)
            brfld.actors_for_sublayer('default')[new_id] = shieldActor
            self.editor.copy_actor_groups({ "actor": old_sName }, { "actor": new_id }, scenario)
            self.editor.remove_entity(reference, None)

            # actually rename it
            shieldActor.sName = new_id
            life_comp[link_name] = self.editor.build_link(new_id)

            # update the minimap entry as well
            mapBlockages = self.editor.get_scenario_map(scenario).raw.Root.mapBlockages
            mapBlockages[new_id] = copy.deepcopy(mapBlockages[old_sName])

            # flip the icon on rightfacing shields in order to optimize the icons file,
            # allowing room for new assets in custom_door_types.py
            if link_name.startswith("wpRight"):
                mapBlockages[new_id]["bFlipX"] = True
            mapBlockages.pop(old_sName)

    def get_shield_id(self, scenario: str):
        # since the available shield ids is auto sorted, just pop the first value
        return f"RandoShield_{self.available_shield_ids[scenario].pop(0)}"

    def reclaim_old_shield_id(self, sName: str, scenario: str):
        # if it's a RandoShield, reclaim the old id after the underscore
        shieldId = int(sName.split("_")[1]) if "RandoShield" in sName else None
        if shieldId is not None:
            insort(self.available_shield_ids[scenario], shieldId)

    def patch_doorpresence_collision(self):
        # extends the door collider in doorpresence actor to 300x300 to maintain the size of normal doors
        # this looks a bit bad, but it'll do until we figure out how to edit navmeshes
        # (or whatever is storing the intended hitboxes for doors)
        doorpresence = self.editor.get_file("actors/props/doorpresence/collisions/doorpresence.bmscd", Bmscc)
        door_collider = doorpresence.raw.layers[0].entries[0]
        door_collider.data.max = ListContainer([300.0, 320.0])



import json
import os
import typing
from pathlib import Path

import lupa
import pytest

from open_dread_rando.patcher_editor import PatcherEditor

_FAIL_INSTEAD_OF_SKIP = True


def get_env_or_skip(env_name, override_fail: typing.Optional[bool] = None):
    if override_fail is None:
        fail_or_skip = _FAIL_INSTEAD_OF_SKIP
    else:
        fail_or_skip = override_fail
    if env_name not in os.environ:
        if fail_or_skip:
            pytest.fail(f"Missing environment variable {env_name}")
        else:
            pytest.skip(f"Skipped due to missing environment variable {env_name}")
    return os.environ[env_name]


class TestFilesDir:
    def __init__(self, root: Path):
        self.root = root

    def joinpath(self, *paths) -> Path:
        return self.root.joinpath(*paths)

    def read_json(self, *paths) -> typing.Union[dict, list]:
        with self.joinpath(*paths).open() as f:
            return json.load(f)


@pytest.fixture(scope="session")
def dread_path():
    return Path(get_env_or_skip("DREAD_1_0_0_PATH"))


@pytest.fixture(scope="session")
def test_files_dir() -> TestFilesDir:
    return TestFilesDir(Path(__file__).parent.joinpath("test_files"))


@pytest.fixture()
def lua_runtime():
    runtime = lupa.LuaRuntime()
    runtime.execute("""
Game = {}
function Game.ImportLibrary() end
function Game.LogWarn() end
function Game.SetForceSkipCutscenes() end
""")
    return runtime


@pytest.fixture()
def patcher_editor(dread_path):
    return PatcherEditor(dread_path)


def pytest_addoption(parser):
    parser.addoption('--skip-if-missing', action='store_false', dest="fail_if_missing",
                     default=True, help="Skip tests instead of missing, in case any asset is missing")


def pytest_configure(config: pytest.Config):
    global _FAIL_INSTEAD_OF_SKIP
    _FAIL_INSTEAD_OF_SKIP = config.option.fail_if_missing



from construct import Container, ListContainer
from mercury_engine_data_structures.formats import Bmsss

from open_dread_rando.patcher_editor import PatcherEditor


def patch_sprites(editor: PatcherEditor):
    """
    Patches a sprite to the hud tileset for a disconnect symbol

    :param editor: the PatcherEditor
    """
    hud_bmsss = editor.get_file("gui/scripts/sprites_hud_tileset.bmsss", Bmsss)
    # position in bctex is (446, 6) with 59x59 pixels in a 512x512 pixels tile
    cont = Container(
        sID="DISCONNECT",
        oTexUVs=Container(
            v2Offset=ListContainer([446 / 512, 6 / 512]),
            v2Scale=ListContainer([59 / 512, 59 / 512])
        )
    )
    hud_bmsss.raw.Root.mapSpriteSheets.HUD_TILESET.vecItems.append(cont)



from __future__ import annotations

import typing

from mercury_engine_data_structures.formats import Txt

if typing.TYPE_CHECKING:
    from open_dread_rando.patcher_editor import PatcherEditor

# We want to edit all the localization files, not just english
ALL_TEXT_FILES = {
    "eu_dutch.txt",
    "eu_french.txt",
    "eu_german.txt",
    "eu_italian.txt",
    "eu_spanish.txt",
    "japanese.txt",
    "korean.txt",
    "russian.txt",
    "simplified_chinese.txt",
    "traditional_chinese.txt",
    "us_english.txt",
    "us_french.txt",
    "us_spanish.txt"
}


def patch_text(editor: PatcherEditor, key: str, value: str):
    for text_file in ALL_TEXT_FILES:
        text = editor.get_file(f"system/localization/{text_file}", Txt)
        text.strings[key] = value


def get_text(editor: PatcherEditor, key: str, text_file: str = "us_english.txt") -> str:
    text = editor.get_file(f"system/localization/{text_file}", Txt)
    return text.strings[key]


def apply_text_patches(editor: PatcherEditor, patches: dict[str, str]):
    for k, v in patches.items():
        patch_text(editor, k, v)


def patch_hints(editor: PatcherEditor, hints: list[dict]):
    # patch mission log titles
    for lvl in {"AQUA", "CAVE", "MAGMA", "FOREST", "LAB", "SANC", "SHIP"}:
        mlog_key = f"MLOG_ADAM_{lvl}"
        original = get_text(editor, mlog_key)
        new = original.replace("Adam Briefing", "a hint")
        patch_text(editor, mlog_key, new)

    for hint in hints:
        ap_actor = editor.resolve_actor_reference(hint["accesspoint_actor"])
        usable = ap_actor.pComponents.USABLE

        # don't lock any doors
        usable.vDoorsToChange = []
        usable.wpThermalDevice = ""

        # update hint text
        hint_id = f"DIAG_ADAM_{hint['hint_id']}"

        hint_text = hint["text"]

        if isinstance(hint_text, str):
            hint_text = [hint_text]

        pages = {
            f"{hint_id}_PAGE_{i + 1}": hint
            for i, hint in enumerate(hint_text)
        }

        usable.tCaptionList = {
            hint_id: list(pages) or [f"{hint_id}_PAGE_1"]
        }
        apply_text_patches(editor, pages)


_PROJECT_MEMBERS = {
    "Project Leads": [
        "Henrique 'Darkszero' Gemignani",
        "duncathan_salt",
    ],
    "Game Patching": [
        "stuckpixel",
        "ScorelessPine",
        "Arcanox",
        "Migs",
        "hyperbola0",
        "Thanatos",
        "Mayberry",
    ],
    "Logic Database": [
        "KirbymastaH",
        "Dyceron",
        "XenoWars",
        "Hugoshido",
        "Tyranisaur",
        "Mayberry",
        "Arcanox",
    ],
    "Additional Art": [
        "BigSharksZ",
        "SkyTheLucario",
        "tezar tantular",
        "hyperbola0",
    ],
    "     ": [
        "With contributions from many others.",
    ]
}


def patch_credits(editor: PatcherEditor, spoiler_log: dict[str, str]):
    text = editor.get_file("system/localization/credits.txt", Txt)
    ordered_credits = list(text.strings.items())

    rando_credits = [
        ("Randomizer Credits", "_TITLE"),
    ]

    for group, group_members in _PROJECT_MEMBERS.items():
        rando_credits.append((group, "_SUBTITLE"))
        for member in group_members:
            rando_credits.append((member, ""))

    if spoiler_log:
        rando_credits.append(("     ", "_SUBTITLE"))
        rando_credits.append(("Major Item Locations", "_TITLE"))
        for item, loc in spoiler_log.items():
            rando_credits.append((item, "_SUBTITLE"))
            rando_credits.append((loc, ""))

    rando_credits.append(("     ", "_SUBTITLE"))

    ordered_credits[1:1] = [
        (f"CREDIT_R_{i:03}{prefix}", item)
        for i, (item, prefix) in enumerate(rando_credits)
    ]
    text.strings = {k: v for k, v in ordered_credits}



import json
import logging
import shutil
import typing
from pathlib import Path

import open_dread_rando_exlaunch
from construct import ListContainer
from mercury_engine_data_structures.file_tree_editor import OutputFormat

from open_dread_rando.constants import FadeTimes
from open_dread_rando.cosmetic_patches import apply_cosmetic_patches
from open_dread_rando.cosmetic_patches.missile_color_patcher import generate_missile_colors
from open_dread_rando.door_locks.custom_door_types import create_all_shield_assets
from open_dread_rando.door_locks.door_patcher import DoorPatcher
from open_dread_rando.files import files_path
from open_dread_rando.logger import LOG
from open_dread_rando.misc_patches import elevator, lua_util
from open_dread_rando.misc_patches.actor_patcher import apply_actor_patches
from open_dread_rando.misc_patches.exefs import patch_exefs
from open_dread_rando.misc_patches.model_patcher import generate_custom_models
from open_dread_rando.misc_patches.sprite_patches import patch_sprites
from open_dread_rando.misc_patches.text_patches import apply_text_patches, patch_credits, patch_hints, patch_text
from open_dread_rando.misc_patches.tilegroup_patcher import patch_individual_tiles, patch_tilegroup
from open_dread_rando.misc_patches.tunable_patcher import apply_tunable_patches
from open_dread_rando.output_config import output_format_for_category, output_paths_for_compatibility
from open_dread_rando.patcher_editor import PatcherEditor
from open_dread_rando.pickups.lua_editor import LuaEditor
from open_dread_rando.pickups.pickup import pickup_object_for
from open_dread_rando.pickups.split_pickups import patch_split_pickups, update_starting_inventory_split_pickups
from open_dread_rando.specific_patches import game_patches
from open_dread_rando.specific_patches.environmental_damage import apply_constant_damage
from open_dread_rando.specific_patches.mass_delete_actors import mass_delete_actors
from open_dread_rando.specific_patches.objective import apply_objective_patches
from open_dread_rando.specific_patches.static_fixes import apply_static_fixes
from open_dread_rando.validator_with_default import DefaultValidatingDraft7Validator

T = typing.TypeVar("T")


def _read_schema():
    with files_path().joinpath("schema.json").open() as f:
        return json.load(f)


def create_custom_init(editor: PatcherEditor, configuration: dict) -> str:
    cosmetic_options: dict = configuration["cosmetic_patches"]["lua"]["custom_init"]
    inventory: dict[str, int] = configuration["starting_items"]
    starting_location: dict = configuration["starting_location"]
    starting_text: list[list[str]] = configuration.get("starting_text", [])
    configuration_identifier: str = configuration["configuration_identifier"]
    enable_remote_lua: bool = configuration.get("enable_remote_lua", False)

    energy_per_tank = configuration["energy_per_tank"]
    energy_per_part = energy_per_tank / 4

    max_life = energy_per_tank - 1

    # increase starting HP if starting with etanks/parts
    if "ITEM_ENERGY_TANKS" in inventory:
        etanks = inventory.pop("ITEM_ENERGY_TANKS")
        max_life += etanks * energy_per_tank
    if "ITEM_LIFE_SHARDS" in inventory and inventory["ITEM_LIFE_SHARDS"] > 0:
        total_shards = inventory["ITEM_LIFE_SHARDS"]
        leftover_shards = total_shards % 4
        if configuration["immediate_energy_parts"] or leftover_shards == 0:
            max_life += total_shards * energy_per_part
            inventory.pop("ITEM_LIFE_SHARDS")
        elif total_shards > leftover_shards:
            max_life += (total_shards - leftover_shards) * energy_per_part
            inventory["ITEM_LIFE_SHARDS"] = leftover_shards

    inventory.update({
        # TODO: expose shuffling these
        "ITEM_WEAPON_POWER_BEAM": 1,
        "ITEM_WEAPON_MISSILE_LAUNCHER": 1,
    })
    inventory = update_starting_inventory_split_pickups(inventory)

    # Game doesn't like to start if some fields are missing, like ITEM_WEAPON_POWER_BOMB_MAX
    final_inventory = {
        "ITEM_MAX_LIFE": max_life,
        "ITEM_CURRENT_SPECIAL_ENERGY": 1000,
        "ITEM_MAX_SPECIAL_ENERGY": 1000,
        "ITEM_METROID_COUNT": 0,
        "ITEM_METROID_TOTAL_COUNT": 40,
        "ITEM_WEAPON_MISSILE_MAX": 0,
        "ITEM_WEAPON_POWER_BOMB_MAX": 0,
    }
    final_inventory.update(inventory)

    def chunks(array, n):
        for i in range(0, len(array), n):
            yield array[i:i + n]

    textboxes = 0
    for group in starting_text:
        boxes = chunks(group, 3)
        for box in boxes:
            textboxes += 1
            box_text = "|".join(box)
            patch_text(editor, f"RANDO_STARTING_TEXT_{textboxes}", box_text)

    layout_uuid = None
    if "layout_uuid" in configuration:
        layout_uuid = lua_util.wrap_string(configuration["layout_uuid"])

    replacement = {
        "enable_remote_lua": enable_remote_lua,
        "new_game_inventory": final_inventory,
        "starting_scenario": lua_util.wrap_string(starting_location["scenario"]),
        "starting_actor": lua_util.wrap_string(starting_location["actor"]),
        "textbox_count": textboxes,
        "energy_per_tank": energy_per_tank,
        "energy_per_part": energy_per_part,
        "immediate_energy_parts": configuration["immediate_energy_parts"],
        "default_x_released": configuration.get("game_patches", {}).get("default_x_released", False),
        "linear_damage_runs": configuration.get("linear_damage_runs"),
        "linear_dps": configuration.get("linear_dps"),
        "configuration_identifier": lua_util.wrap_string(configuration_identifier),
        "required_artifacts": configuration["objective"]["required_artifacts"],
        "enable_death_counter": cosmetic_options["enable_death_counter"],
        "enable_room_ids": False if cosmetic_options["enable_room_name_display"] == "NEVER" else True,
        "room_id_fade_time": FadeTimes.NO_FADE.value if (
            cosmetic_options["enable_room_name_display"] != "WITH_FADE"
            ) else FadeTimes.ROOM_FADE.value,
        "layout_uuid": layout_uuid,
    }

    replacement.update(configuration.get("game_patches", {}))

    return lua_util.replace_lua_template("custom_init.lua", replacement)


def create_collision_camera_table(editor: PatcherEditor, configuration: dict):
    py_dict: dict = configuration["cosmetic_patches"]["lua"]["camera_names_dict"]

    file = lua_util.replace_lua_template("cc_to_room_name.lua", { "room_dict" : py_dict}, True).encode("ascii")
    editor.add_new_asset("system/scripts/cc_to_room_name.lc", file, ["packs/system/system.pkg"])


def patch_pickups(editor: PatcherEditor, lua_scripts: LuaEditor, pickups_config: list[dict], configuration: dict):
    patch_split_pickups(editor)

    # add to the TOC
    editor.add_new_asset("actors/items/randomizer_powerup/scripts/randomizer_powerup.lc", b'', [])

    for i, pickup in enumerate(pickups_config):
        LOG.debug("Writing pickup %d: %s", i, pickup["resources"][0][0]["item_id"])
        try:
            pickup_object_for(lua_scripts, pickup, i, configuration).patch(editor)
        except NotImplementedError as e:
            LOG.warning(e)


def patch_doors(editor: PatcherEditor, doors_config: list[dict], shield_model_config: dict[str, str]):
    editor.map_icon_editor.add_all_new_door_icons()
    create_all_shield_assets(editor, shield_model_config)

    door_patcher = DoorPatcher(editor)
    for door in doors_config:
        door_patcher.patch_door(door["actor"], door["door_type"])


def patch_spawn_points(editor: PatcherEditor, spawn_config: list[dict]):
    # create custom spawn point
    _EXAMPLE_SP = {"scenario": "s010_cave", "actor": "StartPoint0"}
    base_actor = editor.resolve_actor_reference(_EXAMPLE_SP)
    for new_spawn in spawn_config:
        scenario_name = new_spawn["new_actor"]["scenario"]
        new_actor_name = new_spawn["new_actor"]["actor"]
        collision_camera_name = "eg_" + new_spawn["collision_camera_name"]
        new_spawn_pos = ListContainer(
            (new_spawn["location"]["x"], new_spawn["location"]["y"], new_spawn["location"]["z"]))

        scenario = editor.get_scenario(scenario_name)

        editor.copy_actor(scenario_name, new_spawn_pos, base_actor, new_actor_name)
        scenario.add_actor_to_actor_groups(collision_camera_name, new_actor_name)


def add_custom_files(editor: PatcherEditor):
    custom_romfs = files_path().joinpath("romfs")
    for child in custom_romfs.rglob("*"):
        if not child.is_file():
            continue
        relative = child.relative_to(custom_romfs).as_posix()
        logging.info("Adding custom asset %s", relative)
        editor.add_new_asset(str(relative), child.read_bytes(), [])

    lua_libraries = files_path().joinpath("lua_libraries")
    for child in lua_libraries.rglob("*"):
        if not child.is_file():
            continue
        relative = child.relative_to(lua_libraries)
        full_path = Path("system/scripts").joinpath(relative)
        if full_path.suffix == ".lua":
            full_path = full_path.with_suffix(".lc")
        editor.add_new_asset(full_path.as_posix(), child.read_bytes(), ["packs/system/system.pkg"])


def validate(configuration: dict):
    DefaultValidatingDraft7Validator(_read_schema()).validate(configuration)


def apply_patches(editor: PatcherEditor, lua_editor: LuaEditor, configuration: dict):
    # Copy custom files
    add_custom_files(editor)
    generate_missile_colors(editor)
    generate_custom_models(editor)

    # Apply fixes
    apply_static_fixes(editor)

    # Update init.lc
    lua_util.create_script_copy(editor, "system/scripts/init")
    editor.replace_asset(
        "system/scripts/init.lc",
        create_custom_init(editor, configuration).encode("ascii"),
    )

    # Update cc_to_room_name.lua
    create_collision_camera_table(editor, configuration)

    # Update scenario.lc
    lua_util.replace_script(editor, "system/scripts/scenario", "custom_scenario.lua")

    # Update msemenu_mainmenu
    lua_util.replace_script(editor, "gui/scripts/msemenu_mainmenu", "msemenu_mainmenu.lua")

    # Elevators
    if "elevators" in configuration:
        elevator.patch_elevators(editor, configuration["elevators"])

    # Pickups
    patch_pickups(editor, lua_editor, configuration["pickups"], configuration)

    # Hints
    if "hints" in configuration:
        patch_hints(editor, configuration["hints"])

    # Doors
    patch_doors(editor, configuration["door_patches"], configuration["cosmetic_patches"]["shield_versions"])

    # custom spawn points
    patch_spawn_points(editor, configuration["new_spawn_points"])

    for tile_group in configuration["tile_group_patches"]:
        if "tiletype" in tile_group:
            patch_tilegroup(editor, tile_group)
        else:
            patch_individual_tiles(editor, tile_group["actor"], tile_group["tiles"])

    # Text patches
    if "text_patches" in configuration:
        apply_text_patches(editor, configuration["text_patches"])
    patch_credits(editor, configuration["spoiler_log"])

    # Objective
    apply_objective_patches(editor, configuration)

    # Cosmetic patches
    if "cosmetic_patches" in configuration:
        apply_cosmetic_patches(editor, configuration["cosmetic_patches"])

    # Tunable patches
    if "tunables" in configuration:
        apply_tunable_patches(editor, configuration["tunables"])

    # Environmental Damage
    apply_constant_damage(editor, configuration["constant_environment_damage"])

    # Specific game patches
    game_patches.apply_game_patches(editor, configuration.get("game_patches", {}))

    # Mass delete actors
    mass_delete_actors(editor, configuration["mass_delete_actors"])

    # Actor patches
    apply_actor_patches(editor, configuration.get("actor_patches"))

    # remote connector disconnect symbol
    patch_sprites(editor)


def patch_extracted(input_path: Path, output_path: Path, configuration: dict):
    LOG.info("Will patch files from %s", input_path)

    validate(configuration)

    editor = PatcherEditor(input_path)
    lua_editor = LuaEditor()

    # Apply game patches
    apply_patches(editor, lua_editor, configuration)

    out_romfs, out_exefs, exefs_patches = output_paths_for_compatibility(
        output_path,
        configuration["mod_compatibility"],
    )
    shutil.rmtree(out_romfs, ignore_errors=True)
    shutil.rmtree(out_exefs, ignore_errors=True)
    shutil.rmtree(exefs_patches, ignore_errors=True)
    output_format = output_format_for_category(configuration["mod_category"])

    # Exefs
    LOG.info("Creating exefs patches")
    patch_exefs(exefs_patches, configuration)

    if output_format == OutputFormat.ROMFS:
        open_dread_rando_exlaunch.include_depackager(out_exefs)

    LOG.info("Saving modified lua scripts")
    lua_editor.save_modifications(editor)

    LOG.info("Flush modified assets")
    editor.flush_modified_assets()

    if configuration.get("debug_export_modified_files", False):
        editor.save_modified_saves_to(out_romfs.parent.joinpath("_debug"))

    LOG.info("Saving modified pkgs to %s", out_romfs)
    editor.save_modifications(out_romfs, output_format=output_format)

    LOG.info("Done")



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from mercury_engine_data_structures.formats import Bmmdef

from open_dread_rando.constants import ALL_SCENARIOS
from open_dread_rando.misc_patches.text_patches import patch_text

if TYPE_CHECKING:
    from open_dread_rando.patcher_editor import PatcherEditor


@dataclasses.dataclass(frozen=True)
class MapIcon:
    icon_id: str
    coords: tuple[int, int]
    label: str
    disabled_id: str = 'ItemAdquired'
    offset: tuple[int, int] = (0, 0)
    auto_scale: bool = True
    is_global: bool = True
    full_zoom_scale: bool = True
    string_key: str = dataclasses.field(init=False)

    def __post_init__(self):
        object.__setattr__(self, "string_key", f"MAP_ICON_{self.icon_id}")

    def add_to_defs(self, bmmdef: Bmmdef, editor: PatcherEditor):
        patch_text(editor, self.string_key, self.label)
        bmmdef.add_icon(
            self.icon_id,
            self.coords[1],
            self.coords[0],
            f"#{self.string_key}",
            self.disabled_id,
            self.offset,
            self.auto_scale,
            bIsGlobal=self.is_global,
            bFullZoomScale=self.full_zoom_scale
        )


ALL_ICONS: dict[str, MapIcon | str] = {
    "powerup_speedbooster": MapIcon(
        icon_id="ItemSpeedBooster",
        coords=(3, 7),
        label="SPEED BOOSTER"
    ),
    "powerup_widebeam": MapIcon(
        icon_id="ItemWideBeam",
        coords=(0, 6),
        label="WIDE BEAM"
    ),
    "powerup_plasmabeam": MapIcon(
        icon_id="ItemPlasmaBeam",
        coords=(1, 6),
        label="PLASMA BEAM"
    ),
    "powerup_wavebeam": MapIcon(
        icon_id="ItemWaveBeam",
        coords=(2, 6),
        label="WAVE BEAM"
    ),
    "PROGRESSIVE_BEAM": MapIcon(
        icon_id="ItemProgressiveBeam",
        coords=(7, 7),
        label="PROGRESSIVE BEAM"
    ),
    "powerup_chargebeam": MapIcon(
        icon_id="ItemChargeBeam",
        coords=(3, 6),
        label="CHARGE BEAM"
    ),
    "powerup_diffusionbeam": MapIcon(
        icon_id="ItemDiffusionBeam",
        coords=(4, 6),
        label="DIFFUSION BEAM"
    ),
    "PROGRESSIVE_CHARGE": MapIcon(
        icon_id="ItemProgressiveCharge",
        coords=(8, 7),
        label="PROGRESSIVE CHARGE"
    ),
    "powerup_grapplebeam": MapIcon(
        icon_id="ItemGrappleBeam",
        coords=(5, 6),
        label="GRAPPLE BEAM"
    ),
    "powerup_supermissile": MapIcon(
        icon_id="ItemSuperMissile",
        coords=(6, 6),
        label="SUPER MISSILE"
    ),
    "powerup_icemissile": MapIcon(
        icon_id="ItemIceMissile",
        coords=(7, 6),
        label="ICE MISSILE"
    ),
    "PROGRESSIVE_MISSILE": MapIcon(
        icon_id="ItemProgressiveMissile",
        coords=(9, 7),
        label="PROGRESSIVE MISSILE"
    ),
    "powerup_stormmissile": MapIcon(
        icon_id="ItemStormMissile",
        coords=(8, 6),
        label="STORM MISSILE"
    ),
    "powerup_opticcamo": MapIcon(
        icon_id="ItemOpticCamo",
        coords=(9, 6),
        label="PHANTOM CLOAK"
    ),
    "powerup_ghostaura": MapIcon(
        icon_id="ItemGhostAura",
        coords=(10, 6),
        label="FLASH SHIFT"
    ),
    "powerup_sonar": MapIcon(
        icon_id="ItemSonar",
        coords=(11, 6),
        label="PULSE RADAR"
    ),
    "powerup_variasuit": MapIcon(
        icon_id="ItemVariaSuit",
        coords=(12, 6),
        label="VARIA SUIT"
    ),
    "powerup_gravitysuit": MapIcon(
        icon_id="ItemGravitySuit",
        coords=(13, 6),
        label="GRAVITY SUIT"
    ),
    "PROGRESSIVE_SUIT": MapIcon(
        icon_id="ItemProgressiveSuit",
        coords=(10, 7),
        label="PROGRESSIVE SUIT"
    ),
    "powerup_morphball": MapIcon(
        icon_id="ItemMorphBall",
        coords=(14, 6),
        label="MORPH BALL"
    ),
    "powerup_bomb": MapIcon(
        icon_id="ItemBomb",
        coords=(15, 6),
        label="BOMB"
    ),
    "powerup_crossbomb": MapIcon(
        icon_id="ItemCrossBomb",
        coords=(0, 7),
        label="CROSS BOMB"
    ),
    "PROGRESSIVE_BOMB": MapIcon(
        icon_id="ItemProgressiveBomb",
        coords=(11, 7),
        label="PROGRESSIVE BOMB"
    ),
    "powerup_powerbomb": MapIcon(
        icon_id="ItemPowerBomb",
        coords=(1, 7),
        label="MAIN POWER BOMB"
    ),
    "powerup_doublejump": MapIcon(
        icon_id="ItemDoubleJump",
        coords=(4, 7),
        label="SPIN BOOST"
    ),
    "powerup_spacejump": MapIcon(
        icon_id="ItemSpaceJump",
        coords=(5, 7),
        label="SPACE JUMP"
    ),
    "PROGRESSIVE_SPIN": MapIcon(
        icon_id="ItemProgressiveSpin",
        coords=(12, 7),
        label="PROGRESSIVE SPIN"
    ),
    "powerup_spidermagnet": MapIcon(
        icon_id="ItemSpiderMagnet",
        coords=(2, 7),
        label="SPIDER MAGNET"
    ),
    "powerup_screwattack": MapIcon(
        icon_id="ItemScrewAttack",
        coords=(6, 7),
        label="SCREW ATTACK"
    ),
    "item_energytank": MapIcon(
        icon_id="ItemEnergyTankR",
        coords=(5, 0),
        label="ENERGY TANK",
        is_global=False,
        full_zoom_scale=False
    ),
    "item_energyfragment": MapIcon(
        icon_id="ItemEnergyFragmentR",
        coords=(6, 0),
        label="ENERGY PART",
        is_global=False,
        full_zoom_scale=False
    ),
    "item_missiletank": MapIcon(
        icon_id="ItemMissileTankR",
        coords=(7, 0),
        label="MISSILE TANK",
        is_global=False,
        full_zoom_scale=False
    ),
    "item_missiletankplus": MapIcon(
        icon_id="ItemMissileTankPlusR",
        coords=(8, 0),
        label="MISSILE TANK+",
        is_global=False,
        full_zoom_scale=False
    ),
    "item_powerbombtank": MapIcon(
        icon_id="ItemPowerbombExpansionR",
        coords=(9, 0),
        label="POWER BOMB TANK",
        is_global=False,
        full_zoom_scale=False
    ),
    "item_flashshiftupgrade": MapIcon(
        icon_id="ItemFlashShiftUpgrade",
        coords=(15, 5),
        label="FLASH SHIFT UPGRADE",
        is_global=False,
        full_zoom_scale=False
    ),
    "item_speedboostupgrade": MapIcon(
        icon_id="ItemSpeedBoostUpgrade",
        coords=(14, 5),
        label="SPEED BOOSTER UPGRADE",
        is_global=False,
        full_zoom_scale=False
    ),
    "itemsphere": MapIcon(
        icon_id="ItemNothing",
        coords=(14, 7),
        label="NOTHING",
        is_global=False,
        full_zoom_scale=False
    ),
    "unknown": MapIcon(
        icon_id="ItemUnknown",
        coords=(15, 7),
        label="UNKNOWN ITEM"
    ),
    "BlockageIceL": MapIcon(
        icon_id="BlockageIceL",
        coords=(1,3),
        disabled_id="BlockageDisabledL",
        label="ICE MISSILE COVER",
        offset=(-0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageIceR": MapIcon(
        icon_id="BlockageIceR",
        coords=(1,3),
        disabled_id="BlockageDisabledR",
        label="ICE MISSILE COVER",
        offset=(0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageDiffusionL": MapIcon(
        icon_id="BlockageDiffusionL",
        coords=(3,3),
        disabled_id="BlockageDisabledL",
        label="DIFFUSION BEAM COVER",
        offset=(-0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageDiffusionR": MapIcon(
        icon_id="BlockageDiffusionR",
        coords=(3,3),
        disabled_id="BlockageDisabledR",
        label="DIFFUSION BEAM COVER",
        offset=(0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageStormL": MapIcon(
        icon_id="BlockageStormL",
        coords=(5,3),
        disabled_id="BlockageDisabledL",
        label="STORM MISSILE COVER",
        offset=(-0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageStormR": MapIcon(
        icon_id="BlockageStormR",
        coords=(5,3),
        disabled_id="BlockageDisabledR",
        label="STORM MISSILE COVER",
        offset=(0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageScrewL": MapIcon(
        icon_id="BlockageScrewL",
        coords=(5,1),
        disabled_id="BlockageDisabledL",
        label="SCREW ATTACK COVER",
        offset=(-0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageScrewR": MapIcon(
        icon_id="BlockageScrewR",
        coords=(5,1),
        disabled_id="BlockageDisabledR",
        label="SCREW ATTACK COVER",
        offset=(0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageBombL": MapIcon(
        icon_id="BlockageBombL",
        coords=(7,3),
        disabled_id="BlockageDisabledL",
        label="BOMB COVER",
        offset=(-0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageBombR": MapIcon(
        icon_id="BlockageBombR",
        coords=(7,3),
        disabled_id="BlockageDisabledR",
        label="BOMB COVER",
        offset=(0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageCrossBombL": MapIcon(
        icon_id="BlockageCrossBombL",
        coords=(9,3),
        disabled_id="BlockageDisabledL",
        label="CROSS BOMB COVER",
        offset=(-0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageCrossBombR": MapIcon(
        icon_id="BlockageCrossBombR",
        coords=(9,3),
        disabled_id="BlockageDisabledR",
        label="CROSS BOMB COVER",
        offset=(0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockagePowerBombL": MapIcon(
        icon_id="BlockagePowerBombL",
        coords=(11,3),
        disabled_id="BlockageDisabledL",
        label="POWER BOMB COVER",
        offset=(-0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockagePowerBombR": MapIcon(
        icon_id="BlockagePowerBombR",
        coords=(11,3),
        disabled_id="BlockageDisabledR",
        label="POWER BOMB COVER",
        offset=(0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageClosedL": MapIcon(
        icon_id="BlockageClosedL",
        coords=(5,1),
        disabled_id="BlockageDisabledL",
        label="ACCESS CLOSED",
        offset=(-0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
    "BlockageClosedR": MapIcon(
        icon_id="BlockageClosedR",
        coords=(5,1),
        disabled_id="BlockageDisabledR",
        label="ACCESS CLOSED",
        offset=(0.800000011920929, -0.4000000059604645),
        auto_scale=False
    ),
}
ALL_ICONS.update({f"DNA_{i + 1}": MapIcon(
    icon_id=f"ItemDNA{i + 1}",
    coords=(13, 7),
    label=f"METROID DNA {i + 1}"
) for i in range(12)})


class MapIconEditor:
    def __init__(self, editor: PatcherEditor) -> None:
        self.editor = editor
        self.custom_icons = 0
        self.mapdefs = self.editor.get_file("system/minimap/minimap.bmmdef", Bmmdef)

    def _get_icon(self, key: str) -> str:
        icon = ALL_ICONS.get(key, ALL_ICONS["unknown"])
        if isinstance(icon, str):
            return icon
        self.add_icon(icon)
        return icon.icon_id

    def add_icon(self, icon: MapIcon):
        if icon.icon_id not in self.mapdefs.icons:
            icon.add_to_defs(self.mapdefs, self.editor)

    def get_data(self, pickup: dict) -> str:
        if "map_icon" not in pickup:
            return self._get_icon(pickup["model"])

        map_icon: dict = pickup["map_icon"]
        if "icon_id" in map_icon:
            return self._get_icon(map_icon["icon_id"])

        custom_icon: dict = map_icon["custom_icon"]

        if "base_icon" in custom_icon:
            icon = ALL_ICONS.get(custom_icon["base_icon"], ALL_ICONS["unknown"])
            label = custom_icon["label"]
            if "player" in custom_icon:
                label = f"{custom_icon['player'].upper()}'S {icon.label}"
            icon = dataclasses.replace(icon,
                                       icon_id=f"ItemCustom{self.custom_icons}",
                                       label=label
                                       )

        else:
            if "coords" in custom_icon:
                coords = (custom_icon["coords"]["col"], custom_icon["coords"]["row"])
            else:
                coords = (15, 7)
            icon = MapIcon(
                icon_id=f"ItemCustom{self.custom_icons}",
                coords=coords,
                label=custom_icon["label"],
                is_global=custom_icon.get("is_global", True),
                full_zoom_scale=custom_icon.get("full_zoom_scale", True),
                **custom_icon.get("extras", {})
            )

        self.custom_icons += 1
        self.add_icon(icon)
        return icon.icon_id

    def mirror_bmmdef_icons(self):
        # mirrors vanilla shields, grapple and wide boxes, and 2x1 grapples.
        for shield in ["BlockageMissile", "BlockageSuperMissile", "DoorWide", "BlockagePlasma", "BlockageWave",
                       "PropGrappleBox", "PropWideBeamBox", "PropGrappleBlock"]:
            left = self.mapdefs.raw.Root.mapIconDefs[f"{shield}L"]
            right = self.mapdefs.raw.Root.mapIconDefs[f"{shield}R"]
            for field in ["uSpriteRow", "uSpriteCol", "sDisabledIconId"]:
                right[field] = left[field]

            # flip the x-offset on the right one
            right["vAnchorOffset"][0] = -1 * left["vAnchorOffset"][0]

    def mirror_bmmap_icons(self):
        # mirrors props and blockages of mirrored bmmdef icons in each scenario
        props_to_fix = ["PropGrappleBoxR", "PropWideBeamBoxR", "PropGrappleBlockR"]
        blockages_to_fix = ["BlockageMissileR", "BlockageSuperMissileR", "DoorWideR",
                            "BlockagePlasmaR", "BlockageWaveR"]

        for scenario in ALL_SCENARIOS:
            mmap = self.editor.get_scenario_map(scenario)
            props = mmap.get_category("mapProps")
            blockages = mmap.get_category("mapBlockages")

            for sName in props:
                actor = props[sName]
                if actor["sIconId"] in props_to_fix:
                    actor["bFlipX"] = True

            for sName in blockages:
                actor = blockages[sName]
                if actor["sIconId"] in blockages_to_fix:
                    actor["bFlipX"] = True

    def add_all_new_door_icons(self):
        for icon in ["BlockageIceL", "BlockageIceR", "BlockageDiffusionL", "BlockageDiffusionR",
                     "BlockageStormL", "BlockageStormR", "BlockageBombL", "BlockageBombR",
                     "BlockageCrossBombL", "BlockageCrossBombR", "BlockagePowerBombL", "BlockagePowerBombR",
                     "BlockageClosedL", "BlockageClosedR"]:
            self.add_icon(ALL_ICONS[icon])



import logging
import typing
from pathlib import Path

from open_dread_rando.logger import LOG


def patch_with_status_update(input_path: Path, output_path: Path, configuration: dict,
                             status_update: typing.Callable[[float, str], None]):
    from open_dread_rando.dread_patcher import patch_extracted
    total_logs = 108

    class StatusUpdateHandler(logging.Handler):
        count = 0

        def emit(self, record: logging.LogRecord) -> None:
            message = self.format(record)

            # Ignore "Writing <...>.pkg" messages, since there's also Updating...
            if message.startswith("Writing ") and message.endswith(".pkg"):
                return

            # Encoding a bmsad is quick, skip these
            if message.endswith(".bmsad"):
                return

            # These can come up frequently and are benign
            if message.startswith("Skipping extracted file"):
                return

            self.count += 1
            status_update(self.count / total_logs, message)

    new_handler = StatusUpdateHandler()
    # new_handler.setFormatter(logging.Formatter('[%(asctime)s] [%(levelname)s] %(message)s'))
    tree_editor_log = logging.getLogger("mercury_engine_data_structures.file_tree_editor")

    try:
        tree_editor_log.setLevel(logging.DEBUG)
        tree_editor_log.handlers.insert(0, new_handler)
        tree_editor_log.propagate = False
        LOG.setLevel(logging.INFO)
        LOG.handlers.insert(0, new_handler)
        LOG.propagate = False

        patch_extracted(input_path, output_path, configuration)
        if new_handler.count < total_logs:
            status_update(1, f"Done was {new_handler.count}")

    finally:
        tree_editor_log.removeHandler(new_handler)
        tree_editor_log.propagate = True
        LOG.removeHandler(new_handler)
        LOG.propagate = True



from unittest.mock import MagicMock

from mercury_engine_data_structures.formats.brfld import ActorLayer

from open_dread_rando import dread_patcher
from open_dread_rando.specific_patches.mass_delete_actors import mass_delete_actors


def test_cosmetic_options(lua_runtime):
    # Setup
    layoutUUID = "33aead28-ab1f-4d55-bbd9-fc0da5b48469"

    editor = MagicMock()
    configuration = {
        "starting_items": {},
        "starting_location": {
            "scenario": "s010_cave",
            "actor": "StartPoint0"
        },
        "configuration_identifier": "<identifier>",
        "layout_uuid": layoutUUID,
        "cosmetic_patches": {
            "config": {
            },
            "lua": {
                "custom_init": {
                    "enable_death_counter": False,
                    "enable_room_name_display": "NEVER"
                },
                "camera_names_dict": {
                }
            },
            "shield_versions": {
                "ice_missile": "ALTERNATE",
                "diffusion_beam": "ALTERNATE",
                "storm_missile": "ALTERNATE",
                "bomb": "ALTERNATE",
                "cross_bomb": "ALTERNATE",
                "power_bomb": "DEFAULT"
            }
        },
        "energy_per_tank": 75,
        "immediate_energy_parts": True,
        "constant_environment_damage": {
            "heat": 20,
            "cold": 20,
            "lava": 20
        },
        "game_patches": {
            "raven_beak_damage_table_handling": "consistent_low",
            "remove_grapple_blocks_hanubia_shortcut": True,
            "remove_grapple_block_path_to_itorash": True,
            "default_x_released": False,
            "enable_experiment_boss": True,
            "warp_to_start": True,
        },
        "objective": {
            "required_artifacts": 3,
            "hints": []
        },
    }

    # Run
    result = dread_patcher.create_custom_init(editor, configuration)

    # Assert
    lua_runtime.execute("Init = {}")
    lua_runtime.execute(result)

    assert lua_runtime.eval("Init.fEnergyPerTank") == 75
    assert lua_runtime.eval("Init.sLayoutUUID") == layoutUUID

def test_mass_delete_actors(patcher_editor):
    configuration = {
        "to_remove": [
            {
                "scenario": "s020_magma",
                "actor_layer": "rEntitiesLayer",
                "method": "all"
            },
            {
                "scenario": "s010_cave",
                "actor_layer": "rLightsLayer",
                "method": "remove_from_groups",
                "actor_groups": [
                    "lg_collision_camera_001"
                ]
            },
            {
                "scenario": "s030_baselab",
                "actor_layer": "rLightsLayer",
                "method": "keep_from_groups",
                "actor_groups": [
                    "lg_collision_camera_011_Default"
                ]
            }
        ],
        "to_keep": [
            {
                "scenario": "s010_cave",
                "actor_layer": "rLightsLayer",
                "sublayer": "cave_001_light",
                "actor": "spot_001_1"
            }
        ]
    }

    mass_delete_actors(patcher_editor, configuration)

    s010_cave = patcher_editor.get_scenario("s010_cave")
    cave_light_001 = s010_cave.get_actor_group("lg_collision_camera_001", "rLightsLayer")
    cave_spot_001_1_link = patcher_editor.build_link("spot_001_1", "cave_001_light", ActorLayer.LIGHTS)

    assert cave_light_001 == [cave_spot_001_1_link]

    s020_magma = patcher_editor.get_scenario("s020_magma")
    magma_entities = [actor_name for (sublayer_name, actor_name, actor) in s020_magma.all_actors_in_actor_layer()]

    assert len(magma_entities) == 0

    s030_baselab = patcher_editor.get_scenario("s030_baselab")
    lab_light_001 = s030_baselab.get_actor_group("lg_collision_camera_001_Default", "rLightsLayer")
    lab_light_010 = s030_baselab.get_actor_group("lg_collision_camera_010_Default", "rLightsLayer")
    lab_light_011 = s030_baselab.get_actor_group("lg_collision_camera_011_Default", "rLightsLayer")
    cubemap_010_link = patcher_editor.build_link("cubemap_010", "base_010_light", ActorLayer.LIGHTS)

    assert len(lab_light_011) == 3 and len(lab_light_001) == 0 and cubemap_010_link in lab_light_010



import dataclasses
from typing import Optional


@dataclasses.dataclass(frozen=True)
class Transform:
    position: tuple[float, float, float] = (0.0, 0.0, 0.0)
    angle: tuple[float, float, float] = (0.0, 0.0, 0.0)
    scale: tuple[float, float, float] = (1.0, 1.0, 1.0)


@dataclasses.dataclass(frozen=True)
class ModelData:
    bcmdl_path: str
    bmsas: str
    dependencies: tuple[str, ...]
    grapple_fx: bool = False
    transform: Optional[Transform] = None


ALL_MODEL_DATA: dict[str, ModelData] = {
    "itemsphere": ModelData(
        bcmdl_path="actors/items/itemsphere/models/itemsphere.bcmdl",
        bmsas="actors/items/itemsphere/charclasses/timeline.bmsas",
        dependencies=(
            "actors/items/itemsphere/animations/relax.bcskla",
            "actors/items/itemsphere/collisions/itemsphere.bmscd",
            "actors/items/itemsphere/fx/impact.bcptl",
            "actors/items/itemsphere/fx/impact_itemsphere.bcmdl",
            "actors/items/itemsphere/fx/impact_itemsphere.bcskla",
            "actors/items/itemsphere/fx/imats/impact_itemsphere_itemsphere.bsmat",
            "actors/items/itemsphere/models/itemsphere.bcmdl",
            "actors/items/itemsphere/models/imats/itemsphere_mp_opaque_01.bsmat",
        ),
    ),

    "powerup_speedbooster": ModelData(
        bcmdl_path="actors/items/itemsphere/models/speed_booster.bcmdl",
        bmsas="actors/items/itemsphere/charclasses/timeline.bmsas",
        dependencies=(
            "actors/items/itemsphere/animations/relax.bcskla",
            "actors/items/itemsphere/collisions/itemsphere.bmscd",
            "actors/items/itemsphere/fx/impact.bcptl",
            "actors/items/itemsphere/fx/impact_itemsphere.bcmdl",
            "actors/items/itemsphere/fx/impact_itemsphere.bcskla",
            "actors/items/itemsphere/fx/imats/impact_itemsphere_itemsphere.bsmat",
            "actors/items/itemsphere/models/speed_booster.bcmdl",
            "actors/items/itemsphere/models/imats/speedboost_mp_opaque_01.bsmat",
        ),
    ),

    "powerup_widebeam": ModelData(
        bcmdl_path="actors/items/powerup_widebeam/models/powerup_widebeam.bcmdl",
        bmsas="actors/items/powerup_widebeam/charclasses/powerup_widebeam.bmsas",
        dependencies=(
            "actors/items/powerup_widebeam/models/powerup_widebeam.bcmdl",
            "actors/items/powerup_widebeam/models/imats/powerup_widebeam_autoilum.bsmat",
            "actors/items/powerup_widebeam/models/imats/powerup_widebeam_bola.bsmat",
            "actors/items/powerup_widebeam/models/imats/powerup_widebeam_cangrejo.bsmat",
        ),
    ),

    "powerup_wavebeam": ModelData(
        bcmdl_path="actors/items/powerup_widebeam/models/powerup_wavebeam.bcmdl",
        bmsas="actors/items/powerup_widebeam/charclasses/powerup_widebeam.bmsas",
        dependencies=(
            "actors/items/powerup_widebeam/models/powerup_wavebeam.bcmdl",
            "actors/items/powerup_widebeam/models/imats/powerup_wavebeam_autoilum.bsmat",
            "actors/items/powerup_widebeam/models/imats/powerup_wavebeam_bola.bsmat",
            "actors/items/powerup_widebeam/models/imats/powerup_wavebeam_cangrejo.bsmat",
        ),
    ),

    "powerup_plasmabeam": ModelData(
        bcmdl_path="actors/items/powerup_plasmabeam/models/powerup_plasmabeam.bcmdl",
        bmsas="actors/items/powerup_plasmabeam/charclasses/powerup_plasmabeam.bmsas",
        dependencies=(
            "actors/items/powerup_plasmabeam/models/powerup_plasmabeam.bcmdl",
            "actors/items/powerup_plasmabeam/models/imats/powerup_plasmabeam_autoilum.bsmat",
            "actors/items/powerup_plasmabeam/models/imats/powerup_plasmabeam_body.bsmat",
            "actors/items/powerup_plasmabeam/models/imats/powerup_plasmabeam_mp_glass_01.bsmat",
        ),
    ),

    "powerup_chargebeam": ModelData(
        bcmdl_path="actors/items/powerup_chargebeam/models/powerup_chargebeam.bcmdl",
        bmsas="actors/items/powerup_chargebeam/charclasses/powerup_chargebeam.bmsas",
        dependencies=(
            "actors/items/powerup_chargebeam/models/powerup_chargebeam.bcmdl",
            "actors/items/powerup_chargebeam/models/imats/powerup_chargebeam_autoilum.bsmat",
            "actors/items/powerup_chargebeam/models/imats/powerup_chargebeam_body.bsmat",
            "actors/items/powerup_chargebeam/models/imats/powerup_chargebeam_mp_glass_01.bsmat",
        ),
    ),

    "powerup_diffusionbeam": ModelData(
        bcmdl_path="actors/items/powerup_diffusionbeam/models/powerup_diffusionbeam.bcmdl",
        bmsas="actors/items/powerup_diffusionbeam/charclasses/powerup_diffusionbeam.bmsas",
        dependencies=(
            "actors/items/powerup_diffusionbeam/models/powerup_diffusionbeam.bcmdl",
            "actors/items/powerup_diffusionbeam/models/imats/powerup_diffusionbeam_autoilum.bsmat",
            "actors/items/powerup_diffusionbeam/models/imats/powerup_diffusionbeam_mp_glass_01.bsmat",
            "actors/items/powerup_diffusionbeam/models/imats/powerup_diffusionbeam_mp_opaque_01.bsmat",
        ),
    ),

    "powerup_grapplebeam": ModelData(
        bcmdl_path="actors/items/powerup_grapplebeam/models/powerup_grapplebeam.bcmdl",
        bmsas="actors/items/powerup_grapplebeam/charclasses/powerup_grapplebeam.bmsas",
        dependencies=(
            "actors/items/powerup_grapplebeam/fx/auraitemparticle.bcptl",
            "actors/items/powerup_grapplebeam/models/powerup_grapplebeam.bcmdl",
            "actors/items/powerup_grapplebeam/models/imats/powerup_grapplebeam_grapplebeam.bsmat",
        ),
        grapple_fx=True,
    ),

    "powerup_missile": ModelData(
        bcmdl_path="actors/items/powerup_supermissile/models/powerup_missile.bcmdl",
        bmsas="actors/items/powerup_supermissile/charclasses/powerup_supermissile.bmsas",
        dependencies=(
            "actors/items/powerup_supermissile/models/powerup_missile.bcmdl",
            "actors/items/powerup_supermissile/models/imats/powerup_supermissile_hologram.bsmat",
            "actors/items/powerup_supermissile/models/imats/powerup_missile_mp_opaque_01.bsmat",
        ),
    ),

    "powerup_supermissile": ModelData(
        bcmdl_path="actors/items/powerup_supermissile/models/powerup_supermissile.bcmdl",
        bmsas="actors/items/powerup_supermissile/charclasses/powerup_supermissile.bmsas",
        dependencies=(
            "actors/items/powerup_supermissile/models/powerup_supermissile.bcmdl",
            "actors/items/powerup_supermissile/models/imats/powerup_supermissile_hologram.bsmat",
            "actors/items/powerup_supermissile/models/imats/powerup_supermissile_mp_opaque_01.bsmat",
        ),
    ),

    "powerup_icemissile": ModelData(
        bcmdl_path="actors/items/powerup_supermissile/models/powerup_icemissile.bcmdl",
        bmsas="actors/items/powerup_supermissile/charclasses/powerup_supermissile.bmsas",
        dependencies=(
            "actors/items/powerup_supermissile/models/powerup_icemissile.bcmdl",
            "actors/items/powerup_supermissile/models/imats/powerup_supermissile_hologram.bsmat",
            "actors/items/powerup_supermissile/models/imats/powerup_icemissile_mp_opaque_01.bsmat",
        ),
    ),

    "powerup_stormmissile": ModelData(
        bcmdl_path="actors/items/powerup_supermissile/models/powerup_supermissile.bcmdl",
        bmsas="actors/items/powerup_supermissile/charclasses/powerup_supermissile.bmsas",
        dependencies=(
            "actors/items/powerup_supermissile/models/powerup_supermissile.bcmdl",
            "actors/items/powerup_supermissile/models/imats/powerup_supermissile_hologram.bsmat",
            "actors/items/powerup_supermissile/models/imats/powerup_supermissile_mp_opaque_01.bsmat",
            "actors/items/powerup_grapplebeam/fx/auraitemparticle.bcptl",
        ),
        grapple_fx=True,
    ),

    "powerup_opticcamo": ModelData(
        bcmdl_path="actors/items/item_cube_broken/model/itemcube_camo.bcmdl",
        bmsas="actors/items/itemsphere/charclasses/timeline.bmsas",
        dependencies=(
            "actors/items/item_cube_broken/model/itemcube_camo.bcmdl",
            "actors/items/item_cube/model/imats/itemcube_camo.bsmat",
            "actors/items/item_cube/model/imats/itemcube_emisive.bsmat",
        ),
        transform=Transform(
            scale=(0.5, 0.5, 0.5),
            position=(0.0, 30.0, 0.0),
        )
    ),

    "powerup_ghostaura": ModelData(
        bcmdl_path="actors/items/item_cube_broken/model/itemcube_broken.bcmdl",
        bmsas="actors/items/itemsphere/charclasses/timeline.bmsas",
        dependencies=(
            "actors/items/item_cube_broken/model/itemcube_broken.bcmdl",
            "actors/items/item_cube/model/imats/itemcube_cube.bsmat",
            "actors/items/item_cube/model/imats/itemcube_emisive.bsmat",
        ),
        transform=Transform(
            scale=(0.5, 0.5, 0.5),
            position=(0.0, 30.0, 0.0),
        )
    ),

    "powerup_sonar": ModelData(
        bcmdl_path="actors/items/item_cube_broken/model/itemcube_sonr.bcmdl",
        bmsas="actors/items/itemsphere/charclasses/timeline.bmsas",
        dependencies=(
            "actors/items/item_cube_broken/model/itemcube_sonr.bcmdl",
            "actors/items/item_cube/model/imats/itemcube_sonr.bsmat",
            "actors/items/item_cube/model/imats/itemcube_emisive.bsmat",
        ),
        transform=Transform(
            scale=(0.5, 0.5, 0.5),
            position=(0.0, 30.0, 0.0),
        )
    ),

    "powerup_variasuit": ModelData(
        bcmdl_path="actors/items/powerup_variasuit/models/powerup_variasuit.bcmdl",
        bmsas="actors/items/powerup_variasuit/charclasses/powerup_variasuit.bmsas",
        dependencies=(
            "actors/items/powerup_variasuit/models/powerup_variasuit.bcmdl",
            "actors/items/powerup_variasuit/models/imats/powerup_variasuit_material01.bsmat",
        ),
    ),

    "powerup_gravitysuit": ModelData(
        bcmdl_path="actors/items/powerup_gravitysuit/models/powerup_gravitysuit.bcmdl",
        bmsas="actors/items/powerup_gravitysuit/charclasses/powerup_gravitysuit.bmsas",
        dependencies=(
            "actors/items/powerup_gravitysuit/models/powerup_gravitysuit.bcmdl",
            "actors/items/powerup_gravitysuit/models/imats/powerup_gravitysuit_ball.bsmat",
            "actors/items/powerup_gravitysuit/models/imats/powerup_gravitysuit_grav.bsmat",
            "actors/items/powerup_gravitysuit/models/imats/powerup_gravitysuit_mp_opaque_01.bsmat",
        ),
    ),

    "powerup_morphball": ModelData(
        bcmdl_path="actors/items/powerup_bomb/models/powerup_morph.bcmdl",
        bmsas="actors/items/powerup_bomb/charclasses/powerup_bomb.bmsas",
        dependencies=(
            "actors/items/powerup_bomb/models/powerup_morph.bcmdl",
            "actors/items/powerup_bomb/models/imats/powerup_morphhologram.bsmat",
            "actors/items/powerup_bomb/models/imats/powerup_morphmaterial.bsmat",
        ),
        transform=Transform(
            scale=(1.6, 1.6, 1.6),
            position=(0.0, -20.0, 0.0),
        )
    ),

    "powerup_bomb": ModelData(
        bcmdl_path="actors/items/powerup_bomb/models/powerup_bomb.bcmdl",
        bmsas="actors/items/powerup_bomb/charclasses/powerup_bomb.bmsas",
        dependencies=(
            "actors/items/powerup_bomb/models/powerup_bomb.bcmdl",
            "actors/items/powerup_bomb/models/imats/powerup_bomb_hologram.bsmat",
            "actors/items/powerup_bomb/models/imats/powerup_bomb_material.bsmat",
        ),
    ),

    "powerup_crossbomb": ModelData(
        bcmdl_path="actors/items/powerup_bomb/models/powerup_bomb.bcmdl",
        bmsas="actors/items/powerup_bomb/charclasses/powerup_bomb.bmsas",
        dependencies=(
            "actors/items/powerup_bomb/models/powerup_bomb.bcmdl",
            "actors/items/powerup_bomb/models/imats/powerup_bomb_hologram.bsmat",
            "actors/items/powerup_bomb/models/imats/powerup_bomb_material.bsmat",
            "actors/items/powerup_grapplebeam/fx/auraitemparticle.bcptl",
        ),
        grapple_fx=True,
    ),

    "powerup_doublejump": ModelData(
        bcmdl_path="actors/items/powerup_doublejump/models/powerup_doublejump.bcmdl",
        bmsas="actors/items/powerup_doublejump/charclasses/powerup_doublejump.bmsas",
        dependencies=(
            "actors/items/powerup_doublejump/models/powerup_doublejump.bcmdl",
            "actors/items/powerup_doublejump/models/imats/powerup_doublejump_mp_glass_01.bsmat",
            "actors/items/powerup_doublejump/models/imats/powerup_doublejump_mp_opaque_01.bsmat",
        ),
    ),

    "powerup_spidermagnet": ModelData(
        bcmdl_path="actors/items/powerup_doublejump/models/powerup_magnet.bcmdl",
        bmsas="actors/items/powerup_doublejump/charclasses/powerup_doublejump.bmsas",
        dependencies=(
            "actors/items/powerup_doublejump/models/powerup_magnet.bcmdl",
            "actors/items/powerup_doublejump/models/imats/powerup_doublejump_mp_magnet01.bsmat",
            "actors/items/powerup_doublejump/models/imats/powerup_doublejump_mp_white__01.bsmat",
        ),
    ),

    "powerup_spacejump": ModelData(
        bcmdl_path="actors/items/powerup_spacejump/models/powerup_spacejump.bcmdl",
        bmsas="actors/items/powerup_spacejump/charclasses/powerup_spacejump.bmsas",
        dependencies=(
            "actors/items/powerup_spacejump/models/powerup_spacejump.bcmdl",
            "actors/items/powerup_spacejump/models/imats/powerup_spacejump_mat01.bsmat",
        ),
    ),

    "powerup_screwattack": ModelData(
        bcmdl_path="actors/items/powerup_screwattack/models/powerup_screwattack.bcmdl",
        bmsas="actors/items/powerup_screwattack/charclasses/powerup_screwattack.bmsas",
        dependencies=(
            "actors/items/powerup_screwattack/models/powerup_screwattack.bcmdl",
            "actors/items/powerup_screwattack/models/imats/powerup_screwattack_mp_opaque_01.bsmat",
        ),
    ),

    "item_energytank": ModelData(
        bcmdl_path="actors/items/item_energytank/models/item_energytank.bcmdl",
        bmsas="actors/items/item_energytank/charclasses/item_energytank.bmsas",
        dependencies=(
            "actors/items/item_energytank/models/item_energytank.bcmdl",
            "actors/items/item_energytank/models/imats/item_energytank_mp_opaque_01.bsmat",
        ),
    ),

    "item_energyfragment": ModelData(
        bcmdl_path="actors/items/item_energyfragment/models/item_energyfragment.bcmdl",
        bmsas="actors/items/item_energyfragment/charclasses/item_energyfragment.bmsas",
        dependencies=(
            "actors/items/item_energyfragment/models/item_energyfragment.bcmdl",
            "actors/items/item_energyfragment/models/imats/item_energyfragment_glow.bsmat",
            "actors/items/item_energyfragment/models/imats/item_energyfragment_inner.bsmat",
            "actors/items/item_energyfragment/models/imats/item_energyfragment_mat01.bsmat",
            "actors/items/item_energyfragment/models/imats/item_energyfragment_mat02.bsmat",
        ),
    ),

    "item_missiletank": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missiletank.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missiletank.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missiletank_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_orange": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__OR.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__OR.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__OR_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_yellow": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__YL.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__YL.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__YL_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_green": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__GN.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__GN.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__GN_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_blue": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__BL.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__BL.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__BL_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_cyan": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__CY.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__CY.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__CY_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_purple": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__PR.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__PR.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__PR_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_pink": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__PK.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__PK.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__PK_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_magenta": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__MG.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__MG.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__MG_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_black": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__BK.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__BK.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__BK_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_white": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__WH.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__WH.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__WH_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletank_gray": ModelData(
        bcmdl_path="actors/items/item_missiletank/models/item_missile__GY.bcmdl",
        bmsas="actors/items/item_missiletank/charclasses/item_missiletank.bmsas",
        dependencies=(
            "actors/items/item_missiletank/models/item_missile__GY.bcmdl",
            "actors/items/item_missiletank/models/imats/item_missiletank_mat01.bsmat",
            "actors/items/item_missiletank/models/imats/item_missile__GY_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_missiletankplus": ModelData(
        bcmdl_path="actors/items/item_missiletankplus/models/item_missiletankplus.bcmdl",
        bmsas="actors/items/item_missiletankplus/charclasses/item_missiletankplus.bmsas",
        dependencies=(
            "actors/items/item_missiletankplus/models/item_missiletankplus.bcmdl",
            "actors/items/item_missiletankplus/models/imats/item_missiletankplus_mat01.bsmat",
            "actors/items/item_missiletankplus/models/imats/item_missiletankplus_mp_fxhologram_01.bsmat",
        ),
    ),

    "super_missile_expansion": ModelData(
        bcmdl_path="actors/items/item_missiletankplus/models/super_missile_tank.bcmdl",
        bmsas="actors/items/item_missiletankplus/charclasses/item_missiletankplus.bmsas",
        dependencies=(
            "actors/items/item_missiletankplus/models/super_missile_tank.bcmdl",
            "actors/items/item_missiletankplus/models/imats/super_missile_tank_mat01.bsmat",
            "actors/items/item_missiletankplus/models/imats/super_missile_tank_mp_fxhologram_01.bsmat",
        ),
    ),

    "item_powerbombtank": ModelData(
        bcmdl_path="actors/items/item_powerbombtank/models/item_powerbombtank.bcmdl",
        bmsas="actors/items/item_powerbombtank/charclasses/item_powerbombtank.bmsas",
        dependencies=(
            "actors/items/item_powerbombtank/models/item_powerbombtank.bcmdl",
            "actors/items/item_powerbombtank/models/imats/item_powerbombtank_mat01.bsmat",
            "actors/items/item_powerbombtank/models/imats/item_powerbombtank_mp_fxhologram_01.bsmat",
        ),
        transform=Transform(
            scale=(0.8, 0.8, 0.8),
        )
    ),

    "item_flashshiftupgrade": ModelData(
        bcmdl_path="actors/items/item_cube_broken/model/itemcube_broken.bcmdl",
        bmsas="actors/items/itemsphere/charclasses/timeline.bmsas",
        dependencies=(
            "actors/items/item_cube_broken/model/itemcube_broken.bcmdl",
            "actors/items/item_cube/model/imats/itemcube_cube.bsmat",
            "actors/items/item_cube/model/imats/itemcube_emisive.bsmat",
        ),
        transform=Transform(
            scale=(0.35, 0.35, 0.35),
            position=(0.0, 30.0, 0.0),
        ),
        grapple_fx=True,
    ),

    "item_speedboostupgrade": ModelData(
        bcmdl_path="actors/items/itemsphere/models/speed_booster.bcmdl",
        bmsas="actors/items/itemsphere/charclasses/timeline.bmsas",
        dependencies=(
            "actors/items/itemsphere/animations/relax.bcskla",
            "actors/items/itemsphere/collisions/itemsphere.bmscd",
            "actors/items/itemsphere/fx/impact.bcptl",
            "actors/items/itemsphere/fx/impact_itemsphere.bcmdl",
            "actors/items/itemsphere/fx/impact_itemsphere.bcskla",
            "actors/items/itemsphere/fx/imats/impact_itemsphere_itemsphere.bsmat",
            "actors/items/itemsphere/models/speed_booster.bcmdl",
            "actors/items/itemsphere/models/imats/speedboost_mp_opaque_01.bsmat",
        ),
        transform=Transform(
            scale=(0.5, 0.5, 0.5),
            position=(0.0, 30.0, 0.0),
        ),
        grapple_fx=True,
    ),

    "powerup_powerbomb": ModelData(
        bcmdl_path="actors/items/powerup_powerbomb/models/powerup_powerbomb.bcmdl",
        bmsas="actors/items/item_powerbombtank/charclasses/item_powerbombtank.bmsas",
        dependencies=(
            "actors/items/item_powerbombtank/models/item_powerbombtank.bcmdl",
            # use more leading zeros because path needs to be the same length as for item_powerbombtank
            "actors/items/powerup_powerbomb/models/imats/powerup_powerbomb_mat0001.bsmat",
            "actors/items/powerup_powerbomb/models/imats/powerup_powerbomb_mp_fxhologram_0001.bsmat",
        ),
        transform=Transform(
            scale=(-1.25, 1.25, 1.25),
            position=(0.0, -30.0, 0.0),
        ),
    ),

    "shield_icemissile": ModelData(
        bcmdl_path="actors/props/shield_icemissile/models/shield_icemissile.bcmdl",
        bmsas="actors/props/doorshieldmissile/charclasses/doorshieldmissile.bmsas",
        dependencies=(
            "actors/props/shield_icemissile/models/shield_icemissile.bcmdl",
            "actors/props/shield_icemissile/models/imats/shield_icemissile_mp_opaque_01.bsmat"
        )
    ),

    "shield_diffusion": ModelData(
        bcmdl_path="actors/props/shield___diffusion/models/shield_diffusion.bcmdl",
        bmsas="actors/props/door_shield_plasma/charclasses/door_shield_plasma.bmsas",
        dependencies=(
            "actors/props/shield___diffusion/models/shield_diffusion.bcmdl",
            "actors/props/shield___diffusion/models/imats/shield_diffusion_matfx.bsmat",
            "actors/props/shield___diffusion/models/imats/shield_diffusion_mp_opaque_01.bsmat"
        )
    ),

    "shield_storm_mssl": ModelData(
        bcmdl_path="actors/props/shield_storm_mssl/models/shield_storm_mssl.bcmdl",
        bmsas="actors/props/doorshieldmissile/charclasses/doorshieldmissile.bmsas",
        dependencies=(
            "actors/props/shield_storm_mssl/models/shield_storm_mssl.bcmdl",
            "actors/props/shield_storm_mssl/models/imats/shield_storm_mssl_mp_opaque_01.bsmat"
        )
    ),

    "shield_bombs_regular__": ModelData(
        bcmdl_path="actors/props/shield_bombs_regular__/models/shield_bombs_regular__.bcmdl",
        bmsas="actors/props/doorshieldsupermissile/charclasses/doorshieldsupermissile.bmsas",
        dependencies=(
            "actors/props/shield_bombs_regular__/models/shield_bombs_regular__.bcmdl",
            "actors/props/shield_bombs_regular__/models/imats/shield_bombs_regular___mp_opaque_01.bsmat"
        )
    ),

    "shield_cross_bomb_____": ModelData(
        bcmdl_path="actors/props/shield_cross_bomb_____/models/shield_cross_bomb_____.bcmdl",
        bmsas="actors/props/doorshieldmissile/charclasses/doorshieldmissile.bmsas",
        dependencies=(
            "actors/props/shield_cross_bomb_____/models/shield_cross_bomb_____.bcmdl",
            "actors/props/shield_cross_bomb_____/models/imats/shield_cross_bomb______mp_opaque_01.bsmat"
        )
    ),

    "doorshieldpowerbomb___": ModelData(
        bcmdl_path="actors/props/doorshieldpowerbomb___/models/doorshieldpowerbomb___.bcmdl",
        bmsas="actors/props/doorshieldsupermissile/charclasses/doorshieldsupermissile.bmsas",
        dependencies=(
            "actors/props/doorshieldpowerbomb___/models/shield_powerbomb.bcmdl",
            "actors/props/doorshieldpowerbomb___/models/imats/doorshieldpowerbomb____matfx.bsmat",
            "actors/props/doorshieldpowerbomb___/models/imats/doorshieldpowerbomb____mp_opaque_01.bsmat"
        )
    ),
}


def get_data(name: str) -> ModelData:
    return ALL_MODEL_DATA.get(name, ALL_MODEL_DATA["itemsphere"])



import functools
from enum import Enum

from construct import Container
from mercury_engine_data_structures.formats import Bmmap, Bmsad

from open_dread_rando.files import templates_path
from open_dread_rando.misc_patches.text_patches import patch_text
from open_dread_rando.patcher_editor import PatcherEditor
from open_dread_rando.pickups import model_data
from open_dread_rando.pickups.lua_editor import LuaEditor

EXPANSION_ITEM_IDS = {
    "ITEM_ENERGY_TANKS",
    "ITEM_LIFE_SHARDS",
    "ITEM_WEAPON_MISSILE_MAX",
    "ITEM_WEAPON_POWER_BOMB_MAX",
    "ITEM_WEAPON_POWER_BOMB",
    "ITEM_UPGRADE_FLASH_SHIFT_CHAIN",

    # For multiworld
    "ITEM_NONE",
}


@functools.cache
def _read_template_powerup() -> bytes:
    return templates_path().joinpath("template_powerup.bmsad").read_bytes()


class PickupType(Enum):
    ACTOR = "actor"
    EMMI = "emmi"
    COREX = "corex"
    CORPIUS = "corpius"
    CUTSCENE = "cutscene"


class BasePickup:
    def __init__(self, lua_editor: LuaEditor, pickup: dict, pickup_id: int, configuration: dict):
        self.lua_editor = lua_editor
        self.pickup = pickup
        self.pickup_id = pickup_id
        self.configuration = configuration

    def patch(self, editor: PatcherEditor) -> None:
        raise NotImplementedError()


class ActorPickup(BasePickup):
    def patch_single_item_pickup(self, bmsad: Bmsad) -> Bmsad:
        pickable = bmsad.components["PICKABLE"]
        script = bmsad.components["SCRIPT"]

        set_custom = pickable.functions[0]
        item_id: str = self.pickup["resources"][0][0]["item_id"]
        quantity: float = self.pickup["resources"][0][0]["quantity"]

        if item_id == "ITEM_ENERGY_TANKS":
            quantity *= self.configuration["energy_per_tank"]
            set_custom.set_param(4, "Full")
            set_custom.set_param(5, "fCurrentLife")
            set_custom.set_param(6, "LIFE")

            item_id = "fMaxLife"

        elif item_id == "ITEM_LIFE_SHARDS" and self.configuration["immediate_energy_parts"]:
            item_id = "ITEM_NONE"

        elif item_id == "ITEM_LIFE_SHARDS" and not self.configuration["immediate_energy_parts"]:
            item_id = "fLifeShards"
            set_custom.set_param(4, "Custom")
            set_custom.set_param(5, "")
            set_custom.set_param(6, "LIFE")
            set_custom.set_param(7, "#GUI_ITEM_ACQUIRED_ENERGY_SHARD")

            fields = pickable.fields
            caption = self.pickup["caption"]
            if caption == "Energy Part acquired.\nCollect 4 to increase energy capacity.":
                # If text is vanilla, then support showing how many parts we had.
                fields.sOnPickEnergyFragment1Caption = "#GUI_TANK_ACQUIRED_ENERGY_FRAGMENT_1"
                fields.sOnPickEnergyFragment2Caption = "#GUI_TANK_ACQUIRED_ENERGY_FRAGMENT_2"
                fields.sOnPickEnergyFragment3Caption = "#GUI_TANK_ACQUIRED_ENERGY_FRAGMENT_3"
                fields.sOnPickEnergyFragmentCompleteCaption = "#GUI_TANK_ACQUIRED_ENERGY_FRAGMENT_COMPLETE"
            else:
                for field_name in ["sOnPickEnergyFragment1Caption", "sOnPickEnergyFragment2Caption",
                                   "sOnPickEnergyFragment3Caption", "sOnPickEnergyFragmentCompleteCaption"]:
                    setattr(fields, field_name, caption)

        elif item_id in {"ITEM_WEAPON_MISSILE_MAX", "ITEM_WEAPON_POWER_BOMB_MAX", "ITEM_WEAPON_POWER_BOMB"}:
            current = item_id.replace("_MAX", "_CURRENT")
            if item_id == current:
                current += "_CURRENT"

            set_custom.set_param(4, "Custom")
            set_custom.set_param(5, current)
            set_custom.set_param(8, "guicallbacks.OnSecondaryGunsFire")
            set_custom.set_param(13, float(quantity))

        elif "ITEM_RANDO_ARTIFACT" in item_id:
            item_id = "ITEM_NONE"

        script.functions[0].set_param(2, self.lua_editor.get_script_class(self.pickup))

        set_custom.set_param(1, item_id)
        set_custom.set_param(2, float(quantity))

        return bmsad

    def patch_multi_item_pickup(self, bmsad: Bmsad) -> Bmsad:
        pickable = bmsad.components["PICKABLE"]
        script = bmsad.components["SCRIPT"]

        item_id: str = "ITEM_NONE"
        first_progression: str = self.pickup["resources"][0][0]["item_id"]
        first_progression_quantity: int = self.pickup["resources"][0][0]["quantity"]
        if (
            first_progression in {"ITEM_WEAPON_WIDE_BEAM", "ITEM_WEAPON_SUPER_MISSILE"}
            and first_progression_quantity > 0
        ):
            # the gun doesn't appear to be selected properly on pickup unless we do this
            item_id = first_progression

        set_custom_params = pickable.functions[0]
        set_custom_params.set_param(1, item_id)

        script.functions[0].set_param(2, self.lua_editor.get_script_class(
            self.pickup, actordef_name=bmsad.name
        ))

        return bmsad

    def patch_model(self, editor: PatcherEditor, model_names: list[str], actor: Container,
                    new_template: Bmsad) -> None:
        if len(model_names) == 1:
            selected_model_data = model_data.get_data(model_names[0])

            # Update used model
            new_template.model_name = selected_model_data.bcmdl_path
            model_updater = new_template.components["MODELUPDATER"]
            model_updater.functions[0].set_param(1, selected_model_data.bcmdl_path)

            # Apply grapple particles
            if selected_model_data.grapple_fx:
                grapple = editor.get_file("actors/items/powerup_grapplebeam/charclasses/powerup_grapplebeam.bmsad",
                                          Bmsad)
                grapple_components = grapple.components
                components = new_template.components
                components["MATERIALFX"] = grapple_components["MATERIALFX"]
                components["FX"] = grapple_components["FX"]
                new_template.components = components

            if selected_model_data.transform is not None:
                model_updater.fields.vInitScale = list(selected_model_data.transform.scale)
                model_updater.fields.vInitPosWorldOffset = list(selected_model_data.transform.position)
                model_updater.fields.vInitAngWorldOffset = list(selected_model_data.transform.angle)

            # Animation/BMSAS
            action_set_refs = list(new_template.action_set_refs)
            action_set_refs[0] = selected_model_data.bmsas
            new_template.action_set_refs = action_set_refs
        else:
            default_model_data = model_data.get_data(model_names[0])
            model_updater = new_template.components["MODELUPDATER"]

            new_template.model_name = default_model_data.bcmdl_path
            model_updater.type = "CMultiModelUpdaterComponent"
            model_updater.fields.dctModels = {
                name: model_data.get_data(name).bcmdl_path
                for name in model_names
            }
            model_updater.functions = []

            action_set_refs = list(new_template.action_set_refs)
            action_set_refs[:0] = [model_data.get_data(name).bmsas for name in model_names]
            new_template.action_set_refs = action_set_refs

            actor.pComponents.MODELUPDATER["@type"] = "CMultiModelUpdaterComponent"
            actor.pComponents.MODELUPDATER.sModelAlias = model_names[0]

    def patch_minimap_icon(self, editor: PatcherEditor, actor: Container) -> None:
        if "map_icon" in self.pickup and "original_actor" in self.pickup["map_icon"]:
            map_actor = self.pickup["map_icon"]["original_actor"]
        else:
            map_actor = self.pickup["pickup_actor"]

        map_def = editor.get_scenario_file(map_actor["scenario"], Bmmap)
        if map_actor["actor"] in map_def.ability_labels:
            map_def.ability_labels.pop(map_actor["actor"])
        if map_actor["actor"] in map_def.items:
            icon = map_def.items.pop(map_actor["actor"])
            icon.sIconId = editor.map_icon_editor.get_data(self.pickup)
            map_def.items[actor.sName] = icon

    def patch(self, editor: PatcherEditor) -> None:
        template_bmsad = _read_template_powerup()

        pickup_actor = self.pickup["pickup_actor"]
        pkgs_for_level = editor.get_level_pkgs(pickup_actor["scenario"])
        actor = editor.resolve_actor_reference(pickup_actor)

        new_template = Bmsad.parse(template_bmsad, target_game=editor.target_game)
        new_template.name = f"randomizer_powerup_{self.pickup_id}"

        # Update model
        model_names: list[str] = self.pickup["model"]
        self.patch_model(editor, model_names, actor, new_template)

        # Update minimap
        self.patch_minimap_icon(editor, actor)

        # Update caption
        pickable = new_template.components["PICKABLE"]
        pickable.fields.sOnPickCaption = self.pickup["caption"]
        pickable.fields.sOnPickTankUnknownCaption = self.pickup["caption"]

        # Update given item
        if len(self.pickup["resources"]) == 1 and len(self.pickup["resources"][0]) == 1:
            new_template = self.patch_single_item_pickup(new_template)
        else:
            new_template = self.patch_multi_item_pickup(new_template)

        new_path = f"actors/items/randomizer_powerup/charclasses/randomizer_powerup_{self.pickup_id}.bmsad"
        editor.add_new_asset(new_path, new_template, in_pkgs=pkgs_for_level)
        actor.oActorDefLink = f"actordef:{new_path}"

        # Powerup is in plain sight (except for the part we're using the sphere model)
        actor.pComponents.pop("LIFE", None)

        actor.bEnabled = True
        actor.pComponents.MODELUPDATER.bWantsEnabled = True

        # Dependencies
        for level_pkg in pkgs_for_level:
            editor.ensure_present(level_pkg, "system/animtrees/base.bmsat")

            for name in model_names:
                selected_model_data = model_data.get_data(name)
                editor.ensure_present(level_pkg, selected_model_data.bmsas)
                for dep in selected_model_data.dependencies:
                    editor.ensure_present(level_pkg, dep)
                if selected_model_data.grapple_fx:
                    # always include Grapple FX particle or the game could crash
                    editor.ensure_present(level_pkg, "actors/items/powerup_grapplebeam/fx/auraitemparticle.bcptl")

            editor.ensure_present(level_pkg, "actors/items/randomizer_powerup/scripts/randomizer_powerup.lc")


class ActorDefPickup(BasePickup):
    def _patch_actordef_pickup_script_help(self, editor: PatcherEditor) -> None:
        return self.lua_editor.patch_actordef_pickup_script(
            editor,
            self.pickup,
            self.pickup["pickup_lua_callback"],
        )

    def _patch_actordef_pickup(self, editor: PatcherEditor, item_id_field: str) -> tuple[str, Bmsad]:
        self._patch_actordef_pickup_script_help(editor)

        bmsad_path: str = self.pickup["pickup_actordef"]
        actordef = editor.get_file(bmsad_path, Bmsad)

        ai_component = actordef.components["AI"]
        setattr(ai_component.fields, item_id_field, "ITEM_NONE")

        patch_text(editor, self.pickup["pickup_string_key"], self.pickup["caption"])

        return bmsad_path, actordef

    def patch(self, editor: PatcherEditor) -> None:
        raise NotImplementedError()


class EmmiPickup(ActorDefPickup):
    def patch(self, editor: PatcherEditor) -> None:
        bmsad_path, actordef = self._patch_actordef_pickup(editor, "sInventoryItemOnKilled")
        editor.replace_asset(bmsad_path, actordef)


class CoreXPickup(ActorDefPickup):
    def patch(self, editor: PatcherEditor) -> None:
        bmsad_path, actordef = self._patch_actordef_pickup(editor, "sInventoryItemOnBigXAbsorbed")
        editor.replace_asset(bmsad_path, actordef)

    def _patch_actordef_pickup_script_help(self, editor: PatcherEditor) -> None:
        return self.lua_editor.patch_corex_pickup_script(
            editor,
            self.pickup,
            self.pickup["pickup_lua_callback"],
        )


class CorpiusPickup(ActorDefPickup):
    def patch(self, editor: PatcherEditor) -> None:
        bmsad_path, actordef = self._patch_actordef_pickup(editor, "sInventoryItemOnKilled")
        if self.pickup["resources"][0][0]["item_id"] not in EXPANSION_ITEM_IDS:
            actordef.components["AI"].fields.bGiveInventoryItemOnDead = True

        editor.replace_asset(bmsad_path, actordef)


class CutscenePickup(BasePickup):
    def patch(self, editor: PatcherEditor) -> None:
        self.lua_editor.patch_actordef_pickup_script(
            editor,
            self.pickup,
            self.pickup["pickup_lua_callback"],
            'GUI.ShowMessage({}, true, "")'.format(repr(self.pickup["caption"]))
        )


_PICKUP_TYPE_TO_CLASS: dict[PickupType, type[BasePickup]] = {
    PickupType.ACTOR: ActorPickup,
    PickupType.EMMI: EmmiPickup,
    PickupType.COREX: CoreXPickup,
    PickupType.CORPIUS: CorpiusPickup,
    PickupType.CUTSCENE: CutscenePickup,
}


def pickup_object_for(lua_scripts: LuaEditor, pickup: dict, pickup_id: int, configuration: dict) -> "BasePickup":
    pickup_type = PickupType(pickup["pickup_type"])
    return _PICKUP_TYPE_TO_CLASS[pickup_type](lua_scripts, pickup, pickup_id, configuration)



from pathlib import Path

import pytest

from open_dread_rando import dread_patcher

configuration_jsons = [x for x in Path(__file__).parent.joinpath("test_files", "patcher_files").glob("*.json")]

@pytest.mark.parametrize("configuration_path", configuration_jsons)
def test_export(dread_path, tmp_path, test_files_dir, configuration_path):
    output_path = tmp_path.joinpath("out")
    configuration = test_files_dir.read_json(configuration_path)

    dread_patcher.patch_extracted(
        input_path=dread_path,
        output_path=output_path,
        configuration=configuration,
    )
    assert len(list(output_path.rglob("*.lc"))) == 33

    exefs_path = output_path.joinpath("DreadRandovania", "exefs")

    exefs_files = sorted(
        f.relative_to(output_path).as_posix()
        for f in exefs_path.iterdir() if f.name != ".gitignore"
    )
    assert exefs_files == [
        "DreadRandovania/exefs/49161D9CCBC15DF944D0B6278A3C446C006B0BE8.ips",
        "DreadRandovania/exefs/646761F643AFEBB379EDD5E6A5151AF2CEF93DC1.ips",
        "DreadRandovania/exefs/main.npdm",
        "DreadRandovania/exefs/subsdk9",
    ]



from __future__ import annotations

import copy
import dataclasses
import itertools
import typing
from enum import Enum
from typing import Generic, TypeVar

from mercury_engine_data_structures.formats.bmsad import ActorDefFunc, ArgAnyType, Bmsad

from open_dread_rando.patcher_editor import PatcherEditor

if typing.TYPE_CHECKING:
    from typing_extensions import Self


@dataclasses.dataclass
class Color3f:
    r: float
    g: float
    b: float


class BeamInput(Enum):
    NONE = 0
    ZR = 1
    L = 2


class MissileInput(Enum):
    NONE = 0
    L = 1


T = TypeVar('T', bound=ArgAnyType)


class CCFunc:
    def __init__(self, raw: ActorDefFunc) -> None:
        self.raw = raw

    def get_param(self, i: int) -> ArgAnyType:
        return self.raw.get_param(i)

    def set_param(self, i: int, value: ArgAnyType) -> None:
        self.raw.set_param(i, value)


class Param(Generic[T]):
    def __init__(self, index: int) -> None:
        self.index = index

    def __get__(self, inst: CCFunc, objtype: type | None = None) -> T:
        return typing.cast(T, inst.get_param(self.index))

    def __set__(self, inst: CCFunc, value: T) -> None:
        inst.set_param(self.index, value)


class AddPrimaryGun(CCFunc):
    name = Param[str](1)
    class_name = Param[str](2)
    subactor_name = Param[str](3)
    autofire_delay = Param[float](4)
    inventory_item = Param[str](5)
    burst_fx = Param[str](6)
    chargeburst_fx = Param[str](7)
    charge_fx = Param[str](8)
    on_fire = Param[str](9)
    charge_sfx = Param[str](10)
    priority = Param[int](12)

    @property
    def input_type(self) -> BeamInput:
        return BeamInput(self.get_param(11))

    @input_type.setter
    def input_type(self, value: BeamInput) -> None:
        self.set_param(11, value.value)

    @property
    def color(self) -> Color3f:
        return Color3f(
            typing.cast(float, self.get_param(13)),
            typing.cast(float, self.get_param(14)),
            typing.cast(float, self.get_param(15)),
        )

    @color.setter
    def color(self, value: Color3f) -> None:
        self.set_param(13, value.r)
        self.set_param(14, value.g)
        self.set_param(15, value.b)


@dataclasses.dataclass(frozen=True)
class BillBoardGroupParams:
    texture: str
    unk1: int
    unk2: int
    unk3: int
    unk4: float


class SetBillBoardGroupParams(CCFunc):
    name = Param[str](1)

    def _get_group(self, i: int) -> BillBoardGroupParams:
        i *= 5
        return BillBoardGroupParams(
            typing.cast(str, self.get_param(2 + i)),
            typing.cast(int, self.get_param(3 + i)),
            typing.cast(int, self.get_param(4 + i)),
            typing.cast(int, self.get_param(5 + i)),
            typing.cast(float, self.get_param(6 + i)),
        )

    def _set_group(self, i: int, value: BillBoardGroupParams) -> None:
        i *= 5
        self.set_param(2 + i, value.texture)
        self.set_param(3 + i, value.unk1)
        self.set_param(4 + i, value.unk2)
        self.set_param(5 + i, value.unk3)
        self.set_param(6 + i, value.unk4)

    @property
    def beam(self) -> BillBoardGroupParams:
        return self._get_group(0)

    @beam.setter
    def beam(self, value: BillBoardGroupParams) -> None:
        self._set_group(0, value)

    @property
    def charge(self) -> BillBoardGroupParams:
        return self._get_group(1)

    @charge.setter
    def charge(self, value: BillBoardGroupParams) -> None:
        self._set_group(1, value)

    @property
    def boost(self) -> BillBoardGroupParams:
        return self._get_group(2)

    @boost.setter
    def boost(self, value: BillBoardGroupParams) -> None:
        self._set_group(2, value)


class SetGunChargeParams(CCFunc):
    name = Param[str](1)


@dataclasses.dataclass
class PrimaryGunFuncs:
    add_primary_gun: AddPrimaryGun
    set_billboard_group_params: SetBillBoardGroupParams | None
    set_gun_charge_params: SetGunChargeParams | None

    @classmethod
    def from_raw(cls,
                 add_primary_gun: ActorDefFunc,
                 set_billboard_group_params: ActorDefFunc | None = None,
                 set_gun_charge_params: ActorDefFunc | None = None
                 ) -> Self:

        gun = AddPrimaryGun(add_primary_gun)

        if set_billboard_group_params is not None:
            set_billboard_group = SetBillBoardGroupParams(set_billboard_group_params)
        else:
            set_billboard_group = None

        if set_gun_charge_params is not None:
            set_gun_charge = SetGunChargeParams(set_gun_charge_params)
        else:
            set_gun_charge = None

        return cls(
            gun,
            set_billboard_group,
            set_gun_charge,
        )

    @property
    def as_list(self) -> list[ActorDefFunc]:
        ret: list[ActorDefFunc] = [self.add_primary_gun.raw]
        if self.set_billboard_group_params is not None:
            ret.append(self.set_billboard_group_params.raw)
        if self.set_gun_charge_params is not None:
            ret.append(self.set_gun_charge_params.raw)
        return ret

    @property
    def name(self) -> str:
        return self.add_primary_gun.name

    @name.setter
    def name(self, value: str) -> None:
        self.add_primary_gun.name = value
        if self.set_billboard_group_params is not None:
            self.set_billboard_group_params.name = value
        if self.set_gun_charge_params is not None:
            self.set_gun_charge_params.name = value


class AddSecondaryGun(CCFunc):
    name = Param[str](1)
    class_name = Param[str](2)
    subactor_name = Param[str](3)
    fire_delay = Param[float](4)

    main_inventory_item = Param[str](5)
    ammo_inventory_item = Param[str](6)
    capacity_inventory_item = Param[str](11)

    burst_fx = Param[str](7)
    on_fire = Param[str](8)
    priority = Param[int](10)

    @property
    def input_type(self) -> MissileInput:
        return MissileInput(self.get_param(9))

    @input_type.setter
    def input_type(self, value: MissileInput) -> None:
        self.set_param(9, value.value)


def select_gun(param1: str, param2: bool) -> ActorDefFunc:
    result = ActorDefFunc.new("SelectGun")
    result.set_param(1, param1)
    result.set_param(2, param2)
    return result


SAMUS_BMSAD_PATH = "actors/characters/samus/charclasses/samus.bmsad"


# ruff: noqa: C901


def update_starting_inventory_split_pickups(inventory: dict) -> dict:
    inventory = copy.copy(inventory)

    power = "ITEM_WEAPON_POWER_BEAM" in inventory
    wide = "ITEM_WEAPON_WIDE_BEAM" in inventory
    plasma = "ITEM_WEAPON_PLASMA_BEAM" in inventory
    wave = "ITEM_WEAPON_WAVE_BEAM" in inventory

    if power:
        if wide:
            inventory["ITEM_WEAPON_SOLO_WIDE_BEAM"] = 1
        if plasma:
            inventory["ITEM_WEAPON_SOLO_PLASMA_BEAM"] = 1
        if wave:
            inventory["ITEM_WEAPON_SOLO_WAVE_BEAM"] = 1
        if wide and plasma:
            inventory["ITEM_WEAPON_WIDE_PLASMA_BEAM"] = 1
        if wide and wave:
            inventory["ITEM_WEAPON_WIDE_WAVE_BEAM"] = 1
        if plasma and wave:
            inventory["ITEM_WEAPON_PLASMA_WAVE_BEAM"] = 1
        if wide and plasma and wave:
            inventory["ITEM_WEAPON_WIDE_PLASMA_WAVE_BEAM"] = 1

    missile = "ITEM_WEAPON_MISSILE_LAUNCHER" in inventory
    supers = "ITEM_WEAPON_SUPER_MISSILE" in inventory
    ice = "ITEM_WEAPON_ICE_MISSILE" in inventory
    storm = "ITEM_MULTILOCKON" in inventory

    if missile:
        if supers:
            inventory["ITEM_WEAPON_SOLO_SUPER_MISSILE"] = 1
        if ice:
            inventory["ITEM_WEAPON_SOLO_ICE_MISSILE"] = 1
        if supers and ice:
            inventory["ITEM_WEAPON_SUPER_ICE_MISSILE"] = 1
        if storm:
            inventory["ITEM_WEAPON_STORM_MISSILE"] = 1

    return inventory


def patch_split_pickups(editor: PatcherEditor) -> None:
    samus = editor.get_file(SAMUS_BMSAD_PATH, Bmsad)

    primaries = _patch_split_beams(editor)
    secondaries = _patch_split_missiles(editor)
    selections = [
        select_gun("PowerBeam", True),
        select_gun("Missile", False),
    ]

    new_funcs: list[ActorDefFunc] = primaries + secondaries + selections
    samus.components["GUN"].functions = new_funcs

    _patch_blast_shields(editor)


def _patch_split_beams(editor: PatcherEditor) -> list[ActorDefFunc]:
    samus = editor.get_file(SAMUS_BMSAD_PATH, Bmsad)
    gun_funcs = samus.components["GUN"].functions

    power = PrimaryGunFuncs.from_raw(
        gun_funcs[0],
        gun_funcs[1],
        gun_funcs[2],
    )
    wide = PrimaryGunFuncs.from_raw(
        gun_funcs[3],
        gun_funcs[4],
        gun_funcs[5],
    )
    plasma = PrimaryGunFuncs.from_raw(
        gun_funcs[6],
        gun_funcs[7],
        gun_funcs[8],
    )
    wave = PrimaryGunFuncs.from_raw(
        gun_funcs[9],
        gun_funcs[10],
        gun_funcs[11],
    )
    hyper = PrimaryGunFuncs.from_raw(
        gun_funcs[12],
        gun_funcs[13],
    )
    grapple = PrimaryGunFuncs.from_raw(
        gun_funcs[14],
    )
    spbgun = PrimaryGunFuncs.from_raw(
        gun_funcs[15],
    )

    power.add_primary_gun.inventory_item = "ITEM_WEAPON_POWER_BEAM"
    power.add_primary_gun.priority = 0
    assert power.set_billboard_group_params is not None

    solo_wide = copy.deepcopy(wide)
    solo_wide.name = "SoloWideBeam"
    solo_wide.add_primary_gun.inventory_item = "ITEM_WEAPON_SOLO_WIDE_BEAM"
    solo_wide.add_primary_gun.priority = 1

    solo_plasma = copy.deepcopy(plasma)
    solo_plasma.name = "SoloPlasmaBeam"

    solo_plasma.add_primary_gun.inventory_item = "ITEM_WEAPON_SOLO_PLASMA_BEAM"
    solo_plasma.add_primary_gun.priority = 2
    solo_plasma.add_primary_gun.on_fire = "OnPowerBeamFire"
    solo_plasma.add_primary_gun.charge_sfx = "weapons/chargedloop_beam.wav"

    assert solo_plasma.set_billboard_group_params is not None
    solo_plasma.set_billboard_group_params.beam = power.set_billboard_group_params.beam
    solo_plasma.set_billboard_group_params.charge = power.set_billboard_group_params.charge
    solo_plasma.set_billboard_group_params.boost = power.set_billboard_group_params.boost

    solo_wave = copy.deepcopy(wave)
    solo_wave.name = "SoloWaveBeam"
    solo_wave.add_primary_gun.inventory_item = "ITEM_WEAPON_SOLO_WAVE_BEAM"
    solo_wave.add_primary_gun.priority = 3
    solo_wave.add_primary_gun.on_fire = "OnPowerBeamFire"
    solo_wave.add_primary_gun.charge_sfx = "weapons/chargedloop_beam.wav"
    solo_wave.add_primary_gun.color = Color3f(0.8, 0.0, 70.0)
    assert solo_wave.set_billboard_group_params is not None
    solo_wave.set_billboard_group_params.beam = power.set_billboard_group_params.beam
    solo_wave.set_billboard_group_params.charge = power.set_billboard_group_params.charge
    solo_wave.set_billboard_group_params.boost = power.set_billboard_group_params.boost

    wide_plasma = copy.deepcopy(plasma)
    wide_plasma.name = "WidePlasmaBeam"
    wide_plasma.add_primary_gun.inventory_item = "ITEM_WEAPON_WIDE_PLASMA_BEAM"
    wide_plasma.add_primary_gun.priority = 4

    wide_wave = copy.deepcopy(wave)
    wide_wave.name = "WideWaveBeam"
    wide_wave.add_primary_gun.inventory_item = "ITEM_WEAPON_WIDE_WAVE_BEAM"
    wide_wave.add_primary_gun.priority = 5
    wide_wave.add_primary_gun.color = Color3f(0.8, 0.0, 70.0)

    plasma_wave = copy.deepcopy(wave)
    plasma_wave.name = "PlasmaWaveBeam"
    plasma_wave.add_primary_gun.inventory_item = "ITEM_WEAPON_PLASMA_WAVE_BEAM"
    plasma_wave.add_primary_gun.priority = 6
    plasma_wave.add_primary_gun.on_fire = "OnPowerBeamFire"
    plasma_wave.add_primary_gun.charge_sfx = "weapons/chargedloop_beam.wav"
    assert plasma_wave.set_billboard_group_params is not None
    plasma_wave.set_billboard_group_params.beam = power.set_billboard_group_params.beam
    plasma_wave.set_billboard_group_params.charge = power.set_billboard_group_params.charge
    plasma_wave.set_billboard_group_params.boost = power.set_billboard_group_params.boost

    full = copy.deepcopy(wave)
    full.name = "AllBeams"
    full.add_primary_gun.inventory_item = "ITEM_WEAPON_WIDE_PLASMA_WAVE_BEAM"
    full.add_primary_gun.priority = 7

    hyper.add_primary_gun.priority = 8

    return list(itertools.chain.from_iterable(gun.as_list for gun in [
        power,
        solo_wide,
        solo_plasma,
        solo_wave,
        wide_plasma,
        wide_wave,
        plasma_wave,
        full,
        hyper,
        grapple,
        spbgun,
    ]))


def _patch_split_missiles(editor: PatcherEditor) -> list[ActorDefFunc]:
    samus = editor.get_file(SAMUS_BMSAD_PATH, Bmsad)
    gun = samus.components["GUN"]

    solo_ice_bmsad = editor.get_parsed_asset("actors/weapons/icemissile/charclasses/icemissile.bmsad",
                                             type_hint=Bmsad)
    solo_ice_bmsad.name = "soloicemissile"
    movement = solo_ice_bmsad.components["MOVEMENT"].fields
    movement.fInitialSpeed = 1250.0
    movement.fTimeInInitialSpeed = 0.2
    movement.fTimeToReachSpeed = 0.6
    editor.add_new_asset(
        "actors/weapons/soloicemissile/charclasses/soloicemissile.bmsad",
        solo_ice_bmsad,
        editor.find_pkgs("actors/weapons/icemissile/charclasses/icemissile.bmsad"),
    )

    samus.sub_actors.append("soloicemissile")

    missile = AddSecondaryGun(gun.functions[16])
    supers = AddSecondaryGun(gun.functions[17])
    ice = AddSecondaryGun(gun.functions[18])
    lockon = AddSecondaryGun(gun.functions[19])
    omega = AddSecondaryGun(gun.functions[20])

    missile.main_inventory_item = "ITEM_WEAPON_MISSILE_LAUNCHER"
    missile.priority = 0

    solo_super = copy.deepcopy(supers)
    solo_super.name = "SoloSuperMissile"
    solo_super.main_inventory_item = "ITEM_WEAPON_SOLO_SUPER_MISSILE"
    solo_super.priority = 1

    solo_ice = copy.deepcopy(ice)
    solo_ice.name = "SoloIceMissile"
    solo_ice.main_inventory_item = "ITEM_WEAPON_SOLO_ICE_MISSILE"
    solo_ice.subactor_name = "soloicemissile"
    solo_ice.fire_delay = 0.27
    solo_ice.priority = 2

    super_ice = copy.deepcopy(ice)
    super_ice.name = "SuperIceMissile"
    super_ice.main_inventory_item = "ITEM_WEAPON_SUPER_ICE_MISSILE"
    super_ice.priority = 3

    lockon.main_inventory_item = "ITEM_WEAPON_STORM_MISSILE"

    omega.main_inventory_item = "ITEM_WEAPON_POWER_BEAM"

    return [gun.raw for gun in [
        missile,
        solo_super,
        solo_ice,
        super_ice,
        lockon,
        omega,
    ]]


def _patch_blast_shields(editor: PatcherEditor) -> None:
    supers = editor.get_file("actors/props/doorshieldsupermissile/charclasses/doorshieldsupermissile.bmsad", Bmsad)
    super_life = supers.components["LIFE"]
    funcs = list(super_life.functions)
    funcs.pop(1)  # AddDamageSource("SUPER_MISSILE")
    funcs.pop(1)  # AddDamageSource("ICE_MISSILE")
    super_life.functions = funcs

    wide = editor.get_file("actors/props/doorwidebeam/charclasses/doorwidebeam.bmsad", Bmsad)
    wide_life = wide.components["LIFE"]
    funcs = list(wide_life.functions)
    funcs.pop(0)  # AddDamageSource("POWER_BEAM")
    funcs.pop(0)  # AddDamageSource("WIDE_BEAM")
    wide_life.functions = funcs

    plasma = editor.get_file("actors/props/door_shield_plasma/charclasses/door_shield_plasma.bmsad", Bmsad)
    plasma_life = plasma.components["LIFE"]
    funcs = list(plasma_life.functions)
    funcs.pop(1)  # AddDamageSource("PLASMA_BEAM")
    plasma_life.functions = funcs



from unittest.mock import MagicMock

from open_dread_rando.misc_patches import lua_util


def test_create_script_copy_exists():
    editor = MagicMock()
    editor.does_asset_exists.return_value = True

    # Run
    lua_util.create_script_copy(editor, "system/scripts/init")

    # Assert
    editor.does_asset_exists.assert_called_once_with("system/scripts/init_original.lc")
    editor.add_new_asset.assert_not_called()


def test_template_replacement(test_files_dir):
    generated_code = lua_util.replace_lua_template("custom_powerup_template.lua", {
        "name": "RandomizerTestPowerup",
        "parent": "RandomizerPowerup",
        "resources": [
            [{ "item_id": lua_util.wrap_string("ITEM_TEST_POWERUP"), "quantity": 1 }],
        ],
    })

    expected_code = test_files_dir.joinpath("randomizer_progressive_expected.lua").read_text()

    assert generated_code == expected_code



from pathlib import Path

import pytest

from open_dread_rando import dread_patcher

configuration_jsons = [x for x in Path(__file__).parent.joinpath("test_files", "patcher_files").glob("*.json")]

@pytest.mark.parametrize("configuration_path", configuration_jsons)
def test_schema_validation(test_files_dir, configuration_path):
    configuration = test_files_dir.read_json(configuration_path)

    dread_patcher.validate(configuration)



from typing import Any

from mercury_engine_data_structures.formats.ini import Ini

from open_dread_rando.patcher_editor import PatcherEditor


def apply_tunable_patches(editor: PatcherEditor, tunable_patches: dict[str, dict[str, Any]]):
    config_file = editor.get_file('config.ini', Ini)

    for section, items in tunable_patches.items():
        for k, v in items.items():
            if section not in config_file.config:
                config_file.config[section] = {}
            config_file.config[section][k] = Ini.parse_option(v)



import re
from typing import Any

from mercury_engine_data_structures.file_tree_editor import FileTreeEditor

from open_dread_rando.files import files_path, templates_path


def replace_lua_template(file: str, replacement: dict[str, Any], wrap_strings: bool = False) -> str:
    code = templates_path().joinpath(file).read_text()
    for key, content in replacement.items():
        # Replace `TEMPLATE("key")`-style replacements
        code = code.replace(f'TEMPLATE("{key}")', lua_convert(content, wrap_strings))
        # Replace `T__key__T`-style replacements
        code = code.replace(f'T__{key}__T', lua_convert(content, wrap_strings))

    unknown_templates = re.findall(r'TEMPLATE\("([^"]+)"\)', code)
    unknown_templates.extend(re.findall(r'T__(\w+)__T', code))

    if unknown_templates:
        raise ValueError("The following templates were left unfulfilled: " + str(unknown_templates))

    return code


def lua_convert(data, wrap_strings: bool = False) -> str:
    if isinstance(data, list):
        return "{\n" + "\n".join(
            f"{lua_convert(item, wrap_strings)},"
            for item in data
        ) + "\n}"
    if isinstance(data, dict):
        return "{\n" + "\n".join(
            f"{key} = {lua_convert(value, wrap_strings)},"
            for key, value in data.items()
        ) + "\n}"
    if isinstance(data, bool):
        return "true" if data else "false"
    if data is None:
        return "nil"
    if wrap_strings and isinstance(data, str):
        return wrap_string(data)
    return str(data)


def wrap_string(data: str) -> str:
    return f'"{data}"'


def create_script_copy(editor: FileTreeEditor, path: str):
    original = path + "_original.lc"
    if not editor.does_asset_exists(original):
        original_lc = editor.get_raw_asset(path + ".lc")
        editor.add_new_asset(
            original,
            original_lc,
            editor.find_pkgs(path + ".lc")
        )


def replace_script(editor: FileTreeEditor, path: str, replacement_path: str):
    create_script_copy(editor, path)
    editor.replace_asset(path + ".lc", files_path().joinpath(replacement_path).read_bytes())



import dataclasses

from open_dread_rando.misc_patches.material_patcher import MaterialData
from open_dread_rando.misc_patches.model_patcher import StaticModelChanger
from open_dread_rando.patcher_editor import PatcherEditor


# model names for pride missiles
@dataclasses.dataclass()
class MissileTankRecolor(StaticModelChanger):
    def __init__(self, name: str, color: tuple[float, float, float, float]):
        self.materials = {
            "mp_fxhologram_01": MaterialData(
                base_mat="actors/items/item_missiletank/models/imats/item_missiletank_mp_fxhologram_01.bsmat",
                new_mat_name=f"{name}_mp_fxhologram_01",
                new_path=f"actors/items/item_missiletank/models/imats/{name}_mp_fxhologram_01.bsmat",
                uniform_params={
                    "vTex0EmissiveColor": color
                }
            )
        }

        self.base_model_path="actors/items/item_missiletank/models/item_missiletank.bcmdl"
        self.new_model_path=f"actors/items/item_missiletank/models/{name}.bcmdl"


MISSILE_TANK_RECOLORS: dict[str, MissileTankRecolor] = {
    "ORANGE": MissileTankRecolor(
        name="item_missile__OR",
        color=(75.0, 10.0, 0.0, 1.0),
    ),
    "YELLOW": MissileTankRecolor(
        name="item_missile__YL",
        color=(30.0, 30.0, 0.0, 1.0),
    ),
    "GREEN": MissileTankRecolor(
        name="item_missile__GN",
        color=(0.0, 30.0, 0.0, 1.0),
    ),
    "BLUE": MissileTankRecolor(
        name="item_missile__BL",
        color=(0.05, 0.5, 75.0, 1.0),
    ),
    "CYAN": MissileTankRecolor(
        name="item_missile__CY",
        color=(0.05, 10.0, 10.0, 1.0),
    ),
    "PURPLE": MissileTankRecolor(
        name="item_missile__PR",
        color=(15.0, 0.5, 70.0, 1.0),
    ),
    "PINK": MissileTankRecolor(
        name="item_missile__PK",
        color=(70.0, 0.5, 7.0, 1.0),
    ),
    "MAGENTA": MissileTankRecolor(
        name="item_missile__MG",
        color=(70.0, 0.5, 70.0, 1.0),
    ),
    "WHITE": MissileTankRecolor(
        name="item_missile__WH",
        color=(30.0, 30.0, 30.0, 1.0),
    ),
    "BLACK": MissileTankRecolor(
        name="item_missile__BK",
        color=(0.0, 0.0, 0.0, 1.0),
    ),
    "GRAY": MissileTankRecolor(
        name="item_missile__GY",
        color=(0.2, 0.2, 0.2, 1.0),
    ),
}


def generate_missile_colors(editor: PatcherEditor):
    for _, recolor in MISSILE_TANK_RECOLORS.items():
        recolor.generate(editor)
